/********************************************************************************
*                                                                               *
*                    M u l t i - L i ne   T e x t   O b j e c t                 *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998,1999,2000 by Jeroen van der Zijp.   All Rights Reserved.   *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXText.cpp,v 1.106 2000/11/09 16:28:30 jeroen Exp $                      *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXCharset.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXDCWindow.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXScrollbar.h"
#include "FXScrollArea.h"
#include "FXText.h"
#include "FXShell.h"



/*
  Notes:
  - Line start array is one longer than number of visible lines.
  - We want both tab translated to spaces as well as tab-stops array.
  - Control characters in the buffer are OK (e.g. ^L)
  - Mouse based selection callbacks not consistent
  - Redraw when cutting text in middle.
  - Text should return 0 for all unhandled keys!
  - Drag cursor should be same as normal one until drag starts!
  - Change of cursor only implies makePositionVisible() if done by user.
  - INSERT should toggle insert mode/overstrike mode.
  - Breaking:
    Soft-hyphen     173  \xAD
    No break space  240  \xF0
  - Buffer layout:
  
    Content  :  A  B  C  .  .  .  .  .  .  .  .  D  E  F  G 
    Position :  0  1  2 			 3  4  5  6    length=7   
    Addresss :  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14    buffersize=7+11-3=15
             		 ^			 ^		   
	     		 |			 |	  
	     	      gapstart=3	       gapend=11       gaplen=11-3=8
    
    The gap is moved around the buffer so newly added text
    can be entered into the gap; when the gap becomes too small,
    the buffer is resized.  This gapped-buffer technique minimizes
    the number of resizes of the buffer, and minimizes the number
    of block moves.
    
    The tail end of the visrows array will look like:
    
    visrow[0]= 0: "Q R S T U V W \n"   
    visrow[1]= 8: "X Y Z"
    visrow[2]=11: <no text>
    visrow[3]=11: <no text>            length = 11
    
    The last legal position is length = 11.
    
  - While resizing window, keep track of a position which should remain visible,
    i.e. toppos=rowStart(position).  The position is changed same as toppos, except during
    resize.
  - When changing text, if we're looking at the tail end of the buffer, avoid jumping
    the top lines when the content hight shrinks.
  - Add undo capability. (First undo will turn mod flag back off).
  - Add incremental search, search/replace, selection search.
  - Style table stuff.
  - Need to allow for one single routine to update style buffer same as text buffer
  - Suggested additional bindings:
    Ctl-F   Find
    Ctl-R   Replace
    Ctl-G   Find again (Shift-Ctl-G Find again backward)
    Ctl-T   Replace again (Shift-Ctl-G Replace again backward)
    Ctl-L   Goto line number
    Ctl-E   Goto selected
    Ctl-I   Indent (shift 1 character right)
    Ctl-U   Unindent (shift 1 character left)
    Ctl-Z   undo
    Ctl-Y   redo
    Ctl-M   Goto matching (Shift-Ctl-M backward)
    Insert  toggle overstrike mode
    Brace matching
  - Maybe put all keyboard bindings into accelerator table.
  - Need to fix some API's to add notify argument.
*/


#define MINSIZE   80                  // Minimum gap size
#define NVISROWS  20                  // Initial visible rows 

#define MOUSE_NONE    0                // None in effect
#define MOUSE_CHARS   1                // Character drag select
#define MOUSE_WORDS   2                // Word drag select
#define MOUSE_LINES   3                // Line drag select
#define MOUSE_SCROLL  4                // Scrolling mode
#define MOUSE_DRAG    5                // DND-drag
#define MOUSE_PASTE   6                // Text paste
#define MOUSE_TRYDRAG 7                // Tentative DND-drag

#define TEXT_MASK   (TEXT_FIXEDWRAP|TEXT_WORDWRAP|TEXT_OVERSTRIKE|TEXT_READONLY)

/*******************************************************************************/


  
// Map
FXDEFMAP(FXText) FXTextMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXText::onPaint),
  FXMAPFUNC(SEL_MOTION,0,FXText::onMotion),
  FXMAPFUNC(SEL_DRAGGED,0,FXText::onDragged),
  FXMAPFUNC(SEL_TIMEOUT,FXWindow::ID_CARETBLINK,FXText::onBlink),
  FXMAPFUNC(SEL_TIMEOUT,FXWindow::ID_AUTOSCROLL,FXText::onAutoScroll),
  FXMAPFUNC(SEL_FOCUSIN,0,FXText::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXText::onFocusOut),
  FXMAPFUNC(SEL_BEGINDRAG,0,FXText::onBeginDrag),
  FXMAPFUNC(SEL_ENDDRAG,0,FXText::onEndDrag),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXText::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXText::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXText::onMiddleBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXText::onMiddleBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXText::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXText::onRightBtnRelease),
  FXMAPFUNC(SEL_UNGRABBED,0,FXText::onUngrabbed),
  FXMAPFUNC(SEL_CHANGED,0,FXText::onChanged),
  FXMAPFUNC(SEL_DND_ENTER,0,FXText::onDNDEnter),
  FXMAPFUNC(SEL_DND_LEAVE,0,FXText::onDNDLeave),
  FXMAPFUNC(SEL_DND_DROP,0,FXText::onDNDDrop),
  FXMAPFUNC(SEL_DND_MOTION,0,FXText::onDNDMotion),
  FXMAPFUNC(SEL_DND_REQUEST,0,FXText::onDNDRequest),
  FXMAPFUNC(SEL_SELECTION_LOST,0,FXText::onSelectionLost),
  FXMAPFUNC(SEL_SELECTION_GAINED,0,FXText::onSelectionGained),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,FXText::onSelectionRequest),
  FXMAPFUNC(SEL_CLIPBOARD_LOST,0,FXText::onClipboardLost),
  FXMAPFUNC(SEL_CLIPBOARD_GAINED,0,FXText::onClipboardGained),
  FXMAPFUNC(SEL_CLIPBOARD_REQUEST,0,FXText::onClipboardRequest),
  FXMAPFUNC(SEL_KEYPRESS,0,FXText::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXText::onKeyRelease),
  FXMAPFUNC(SEL_SELECTED,0,FXText::onSelected),
  FXMAPFUNC(SEL_DESELECTED,0,FXText::onDeselected),
  FXMAPFUNC(SEL_INSERTED,0,FXText::onInserted),
  FXMAPFUNC(SEL_DELETED,0,FXText::onDeleted),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_HELP,FXText::onQueryHelp),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_TOGGLE_EDITABLE,FXText::onUpdToggleEditable),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_TOGGLE_OVERSTRIKE,FXText::onUpdToggleOverstrike),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_CURSOR_ROW,FXText::onUpdCursorRow),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_CURSOR_COLUMN,FXText::onUpdCursorColumn),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_CUT_SEL,FXText::onUpdHaveSelection),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_COPY_SEL,FXText::onUpdHaveSelection),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_PASTE_SEL,FXText::onUpdYes),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_DELETE_SEL,FXText::onUpdHaveSelection),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_TEXT_COLOR,FXText::onUpdTextColor),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_BACK_COLOR,FXText::onUpdBackColor),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_SELTEXT_COLOR,FXText::onUpdSelTextColor),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_SELBACK_COLOR,FXText::onUpdSelBackColor),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_CURSOR_COLOR,FXText::onUpdCursorColor),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_TAB_COLUMNS,FXText::onUpdTabColumns),
  FXMAPFUNC(SEL_UPDATE,FXText::ID_WRAP_COLUMNS,FXText::onUpdWrapColumns),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_TOP,FXText::onCmdCursorTop),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_BOTTOM,FXText::onCmdCursorBottom),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_HOME,FXText::onCmdCursorHome),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_END,FXText::onCmdCursorEnd),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_RIGHT,FXText::onCmdCursorRight),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_LEFT,FXText::onCmdCursorLeft),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_UP,FXText::onCmdCursorUp),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_DOWN,FXText::onCmdCursorDown),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_WORD_LEFT,FXText::onCmdCursorWordLeft),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_WORD_RIGHT,FXText::onCmdCursorWordRight),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_PAGEDOWN,FXText::onCmdCursorPageDown),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_PAGEUP,FXText::onCmdCursorPageUp),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_SCRNTOP,FXText::onCmdCursorScreenTop),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_SCRNBTM,FXText::onCmdCursorScreenBottom),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_SCRNCTR,FXText::onCmdCursorScreenCenter),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_PAR_HOME,FXText::onCmdCursorParHome),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_PAR_END,FXText::onCmdCursorParEnd),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_MARK,FXText::onCmdMark),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_EXTEND,FXText::onCmdExtend),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_INSERT_CHAR,FXText::onCmdInsertChar),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_INSERT_STRING,FXText::onCmdInsertString),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_INSERT_NEWLINE,FXText::onCmdInsertNewline),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CUT_SEL,FXText::onCmdCutSel),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_COPY_SEL,FXText::onCmdCopySel),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_PASTE_SEL,FXText::onCmdPasteSel),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_DELETE_SEL,FXText::onCmdDeleteSel),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_SELECT_CHAR,FXText::onCmdSelectChar),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_SELECT_WORD,FXText::onCmdSelectWord),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_SELECT_LINE,FXText::onCmdSelectLine),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_SELECT_ALL,FXText::onCmdSelectAll),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_DESELECT_ALL,FXText::onCmdDeselectAll),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_BACKSPACE,FXText::onCmdBackspace),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_BACKSPACE_WORD,FXText::onCmdBackspaceWord),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_BACKSPACE_BOL,FXText::onCmdBackspaceBol),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_DELETE,FXText::onCmdDelete),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_DELETE_WORD,FXText::onCmdDeleteWord),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_DELETE_EOL,FXText::onCmdDeleteEol),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_DELETE_LINE,FXText::onCmdDeleteLine),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_TOGGLE_EDITABLE,FXText::onCmdToggleEditable),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_TOGGLE_OVERSTRIKE,FXText::onCmdToggleOverstrike),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_ROW,FXText::onCmdCursorRow),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_COLUMN,FXText::onCmdCursorColumn),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_TEXT_COLOR,FXText::onCmdTextColor),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_BACK_COLOR,FXText::onCmdBackColor),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_SELTEXT_COLOR,FXText::onCmdSelTextColor),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_SELBACK_COLOR,FXText::onCmdSelBackColor),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_CURSOR_COLOR,FXText::onCmdCursorColor),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETSTRINGVALUE,FXText::onCmdSetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETSTRINGVALUE,FXText::onCmdGetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_TAB_COLUMNS,FXText::onCmdTabColumns),
  FXMAPFUNC(SEL_COMMAND,FXText::ID_WRAP_COLUMNS,FXText::onCmdWrapColumns),
  };


// Object implementation
FXIMPLEMENT(FXText,FXScrollArea,FXTextMap,ARRAYNUMBER(FXTextMap))

  

/*******************************************************************************/


// Helper
static inline FXint fxabs(FXint a){ return a<0?-a:a; }


// For deserialization
FXText::FXText(){
  flags|=FLAG_ENABLED;
  nrows=1;
  nvisrows=0;
  gapstart=0;
  gapend=0;
  toppos=0;
  keeppos=0;
  toprow=0;
  selstartpos=0;
  selendpos=0;  
  anchorpos=0;
  cursorpos=0;
  cursorrow=0;
  cursorcol=0;
  prefcol=-1;
  wrapwidth=80;
  wrapcolumns=80;
  tabwidth=8;
  tabcolumns=8;
  blinker=NULL;
  textWidth=0;
  textHeight=0;
  clipbuffer=NULL;
  cliplength=0;
  modified=FALSE;
  dragmode=MOUSE_NONE;
  grabx=0;
  graby=0;
  }


// Text widget
FXText::FXText(FXComposite *p,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXScrollArea(p,opts,x,y,w,h),
  delimiters("~.,/\\`'!@#$%^&*()-=+{}|[]\":;<>?"){  // I *swear*, these are delimiters :-)
  flags|=FLAG_ENABLED;
  target=tgt;
  message=sel;
  FXCALLOC(&buffer,FXchar,MINSIZE);
  FXCALLOC(&visrows,FXint,NVISROWS+1);
  length=0;
  nrows=1;
  nvisrows=NVISROWS;
  gapstart=0;
  gapend=MINSIZE;
  toppos=0;
  keeppos=0;
  toprow=0;
  selstartpos=0;
  selendpos=0;  
  anchorpos=0;
  cursorpos=0;
  cursorrow=0;
  cursorcol=0;
  prefcol=-1;
  margintop=2;
  marginbottom=2;
  marginleft=3;
  marginright=3;
  wrapwidth=80;
  wrapcolumns=80;
  tabwidth=8;
  tabcolumns=8;
  font=getApp()->getNormalFont();
  defaultCursor=getApp()->textCursor;
  dragCursor=getApp()->textCursor;
  textColor=getApp()->foreColor;
  selbackColor=getApp()->selbackColor;
  seltextColor=getApp()->selforeColor;
  cursorColor=FXRGB(192,0,0);
  blinker=NULL;
  textWidth=0;
  textHeight=0;
  clipbuffer=NULL;
  cliplength=0;
  modified=FALSE;
  dragmode=MOUSE_NONE;
  grabx=0;
  graby=0;
  }


// Create window
void FXText::create(){
  FXScrollArea::create();
  font->create();
  if(!deleteType){deleteType=getApp()->registerDragType(deleteTypeName);}
  if(!textType){textType=getApp()->registerDragType(textTypeName);}
  if(options&TEXT_FIXEDWRAP){wrapwidth=wrapcolumns*font->getTextWidth(" ",1);}
  tabwidth=tabcolumns*font->getTextWidth(" ",1);
  dropEnable();
  recalc();// FIXME:- ugly because replaceText resets the flags...
  }


// Detach window
void FXText::detach(){
  FXScrollArea::detach();
  font->detach();
  deleteType=0;
  textType=0;
  }


// If window can have focus
FXbool FXText::canFocus() const { return 1; }


// Make a valid position
FXint FXText::validPos(FXint pos) const {
  return pos<0 ? 0 : pos>length ? length : pos;
  }


// Enable the window
void FXText::enable(){
  if(!(flags&FLAG_ENABLED)){
    FXScrollArea::enable();
    update(0,0,viewport_w,viewport_h);
    }
  }


// Disable the window
void FXText::disable(){
  if(flags&FLAG_ENABLED){
    FXScrollArea::disable();
    update(0,0,viewport_w,viewport_h);
    }
  }


// Propagate size change
void FXText::recalc(){ 
  FXScrollArea::recalc();
  flags|=FLAG_RECALC;
  }


/*******************************************************************************/


// Get character
FXchar FXText::getChar(FXint pos) const {
  FXASSERT(0<=pos && pos<length);
  return buffer[pos<gapstart ? pos : pos-gapstart+gapend];
  }


// Set character
void FXText::setChar(FXint pos,FXchar ch){
  FXASSERT(0<=pos && pos<length);
  buffer[pos<gapstart ? pos : pos-gapstart+gapend]=ch;
  }


// Move the gap
void FXText::movegap(FXint pos){
  register FXint gaplen=gapend-gapstart;
  FXASSERT(0<=pos && pos<=length);
  FXASSERT(0<=gapstart && gapstart<=length);
  if(gapstart<pos){ 
    memmove(&buffer[gapstart],&buffer[gapend],pos-gapstart);
    gapend=pos+gaplen;
    gapstart=pos;
    }
  else if(pos<gapstart){
    memmove(&buffer[pos+gaplen],&buffer[pos],gapstart-pos);
    gapend=pos+gaplen;
    gapstart=pos;
    }
  }


// Size gap
void FXText::sizegap(FXint sz){
  register FXint gaplen=gapend-gapstart;
  FXASSERT(0<=gapstart && gapstart<=length);
  if(sz>=gaplen){
    sz+=MINSIZE;
    if(!FXRESIZE(&buffer,FXchar,length+sz)){ fxerror("%s::sizegap: out of memory.\n",getClassName()); }
    memmove(&buffer[gapstart+sz],&buffer[gapend],length-gapstart);
    gapend=gapstart+sz;
    }
  }


// Squeeze out the gap by moving it to the end of the buffer;
// This leaves an unbroken character array terminated by a null.
void FXText::squeezegap(){
  if(gapstart!=length){
    memmove(&buffer[gapstart],&buffer[gapend],length-gapstart);
    gapend=length+gapend-gapstart;
    gapstart=length;
    }
  }


/*******************************************************************************/


// Character width
FXint FXText::charWidth(FXchar ch,FXint indent) const {
  register FXint w;
  if(' ' <= ((FXuchar)ch)){
    w=font->getTextWidth(&ch,1);
    }
  else if(ch == '\t'){
    w=(tabwidth-indent%tabwidth);
    }
  else{
    ch|=0x40; 
    w=font->getTextWidth("^",1)+font->getTextWidth(&ch,1);
    }
  return w;
  }


// Start of next wrapped line
FXint FXText::wrap(FXint start) const {
  register FXint lw,cw,p,s,c;
  FXASSERT(0<=start && start<=length);
  lw=0;
  p=s=start;
  while(p<length){
    c=getChar(p);
    if(c=='\n') return p+1;     // Newline always breaks
    cw=charWidth(c,lw);
    if(lw+cw>wrapwidth){
      if(s>start) return s;     // We remembered the last space we encountered; break there!
      if(p==start) p++;         // Always at least one character on each line!
      return p;
      }
    lw+=cw;
    p++;
    if(isspace(c)) s=p;         // Remember potential break point!
    }
  return length;
  }


// Count number of newlines
FXint FXText::countLines(FXint start,FXint end) const {
  register FXint p,nl=0;
  FXASSERT(0<=start && end<=length+1);
  p=start;
  while(p<end){ 
    if(p>=length) return nl+1;
    if(getChar(p)=='\n') nl++; 
    p++;
    }
  return nl;
  }


// Count number of rows; start and end should be on a row start
FXint FXText::countRows(FXint start,FXint end) const {
  register FXint p,q,s,w=0,c,cw,nr=0;
  FXASSERT(0<=start && end<=length+1);
  if(options&TEXT_WORDWRAP){
    p=q=s=start;
    while(q<end){
      if(p>=length) return nr+1;
      c=getChar(p);
      if(c=='\n'){
        nr++;
        w=0;
        p=q=s=p+1;
        }
      else{
        cw=charWidth(c,w);
        if(w+cw>wrapwidth){
          nr++;
          if(s>q){
            p=q=s;
            }
          else{
            if(p==q) p++;
            q=s=p;
            }
          w=0;
          }
        else{
          w+=cw;
          p++;
          if(isspace(c)) s=p;
          }
        }
      }
    }
  else{
    p=start;
    while(p<end){ 
      if(p>=length) return nr+1;
      c=getChar(p);
      if(c=='\n') nr++; 
      p++;
      }
    }
  return nr;
  }


// Measure lines; start and end should be on a row start
FXint FXText::measureText(FXint start,FXint end,FXint& wmax,FXint& hmax) const {
  register FXint nr=0,w=0,c,cw,p,q,s;
  FXASSERT(0<=start && end<=length+1);
  if(options&TEXT_WORDWRAP){
    wmax=wrapwidth;
    p=q=s=start;
    while(q<end){
      if(p>=length){ 
        nr++; 
        break; 
        }
      c=getChar(p);
      if(c=='\n'){
        nr++;
        w=0;
        p=q=s=p+1;
        }
      else{
        cw=charWidth(c,w);
        if(w+cw>wrapwidth){
          nr++;
          if(s>q){
            p=q=s;
            }
          else{
            if(p==q) p++;
            q=s=p;
            }
          w=0;
          }
        else{
          w+=cw;
          p++;
          if(isspace(c)) s=p;
          }
        }
      }
    }
  else{
    wmax=0;
    p=start;
    while(p<end){
      if(p>=length){ 
        if(w>wmax) wmax=w; 
        nr++; 
        break; 
        }
      c=getChar(p);
      if(c=='\n'){ 
        if(w>wmax) wmax=w; 
        nr++; 
        w=0; 
        }
      else{
        w+=charWidth(c,w);
        }
      p++;
      }
    }
  hmax=nr*font->getFontHeight();
  return nr;
  }


// Find end of previous word
FXint FXText::leftWord(FXint pos) const {
  register FXchar ch;
  if(pos>length) pos=length;
  if(0<pos){
    ch=getChar(pos-1);
    if(delimiters.has(ch)) return pos-1;
    }
  while(0<pos){
    ch=getChar(pos-1);
    if(delimiters.has(ch)) return pos;
    if(isspace(ch)) break;
    pos--;
    }
  while(0<pos){
    ch=getChar(pos-1);
    if(!isspace(ch)) return pos;
    pos--;
    }
  return 0;
  }


// Find begin of next word
FXint FXText::rightWord(FXint pos) const {
  register FXchar ch;
  if(pos<0) pos=0;
  if(pos<length){
    ch=getChar(pos);
    if(delimiters.has(ch)) return pos+1;
    }
  while(pos<length){
    ch=getChar(pos);
    if(delimiters.has(ch)) return pos;
    if(isspace(ch)) break;
    pos++;
    }
  while(pos<length){
    ch=getChar(pos);
    if(!isspace(ch)) return pos;
    pos++;
    }
  return length;
  }


// Find begin of a word
FXint FXText::wordStart(FXint pos) const {
  register FXchar c=' ';
  if(pos<=0) return 0; 
  if(pos<length) c=getChar(pos);
  if(c==' ' || c=='\t'){
    while(0<pos){
      c=getChar(pos-1);
      if(c!=' ' && c!='\t') return pos;
      pos--;
      }
    }
  else if(delimiters.has(c)){
    while(0<pos){
      c=getChar(pos-1);
      if(!delimiters.has(c)) return pos;
      pos--;
      }
    }
  else{
    while(0<pos){
      c=getChar(pos-1);
      if(delimiters.has(c) || isspace(c)) return pos;
      pos--;
      }
    }
  return 0;
  }


// Find end of word
FXint FXText::wordEnd(FXint pos) const {
  register FXchar c=' ';
  if(pos>=length) return length; 
  c=getChar(pos); 
  if(c==' ' || c=='\t'){
    while(pos<length){
      c=getChar(pos);
      if(c!=' ' && c!='\t') return pos;
      pos++;
      }
    }
  else if(delimiters.has(c)){
    while(pos<length){
      c=getChar(pos);
      if(!delimiters.has(c)) return pos;
      pos++;
      }
    }
  else{
    while(pos<length){
      c=getChar(pos);
      if(delimiters.has(c) || isspace(c)) return pos;
      pos++;
      }
    }
  return length;
  }


// Return position of begin of paragraph 
FXint FXText::lineStart(FXint pos) const {
  FXASSERT(0<=pos && pos<=length);
  while(0<pos){ if(getChar(pos-1)=='\n') return pos; pos--; }
  return 0;
  }


// Return position of end of paragraph
FXint FXText::lineEnd(FXint pos) const {
  FXASSERT(0<=pos && pos<=length);
  while(pos<length){ if(getChar(pos)=='\n') return pos; pos++; }
  return length;
  }


// Return start of next line
FXint FXText::nextLine(FXint pos,FXint nl) const {
  FXASSERT(0<=pos && pos<=length);
  if(nl<=0) return pos;
  while(pos<length){
    if(getChar(pos)=='\n'){
      if(--nl==0) return pos+1;
      }
    pos++;
    }
  return length;
  }


// Return start of previous line
FXint FXText::prevLine(FXint pos,FXint nl) const {
  FXASSERT(0<=pos && pos<=length);
  if(nl<=0) return pos;
  while(0<pos){
    if(getChar(pos-1)=='\n'){
      if(nl--==0) return pos;
      }
    pos--;
    }
  return 0;
  }


// Return row start
FXint FXText::rowStart(FXint pos) const {
  register FXint p,t;
  FXASSERT(0<=pos && pos<=length);
  if(!(options&TEXT_WORDWRAP)) return lineStart(pos);
  p=lineStart(pos);
  while(p<pos && (t=wrap(p))<=pos && t<length) p=t;
  FXASSERT(p<=pos);
  return p;
  }


// Return row end
FXint FXText::rowEnd(FXint pos) const {
  register FXint p;
  FXASSERT(0<=pos && pos<=length);
  if(!(options&TEXT_WORDWRAP)) return lineEnd(pos);
  p=lineStart(pos);
  while(p<length && p<=pos) p=wrap(p);
  if(pos<p && isspace(getChar(p-1))) p--;
  FXASSERT(pos<=p);
  return p;
  }


// Move to next row given start of line
FXint FXText::nextRow(FXint pos,FXint nr) const {
  FXASSERT(0<=pos && pos<=length);
  if(!(options&TEXT_WORDWRAP)) return nextLine(pos,nr);
  if(nr<=0) return pos;
  pos=rowStart(pos);
  while(pos<length && 0<nr){ pos=wrap(pos); nr--; }
  return pos;
  }


// Move to previous row given start of line
FXint FXText::prevRow(FXint pos,FXint nr) const {
  register FXint p,q,t;
  FXASSERT(0<=pos && pos<=length);
  if(!(options&TEXT_WORDWRAP)) return prevLine(pos,nr);
  if(nr<=0) return pos;
  while(0<pos){
    p=lineStart(pos);
    for(q=p; q<pos && (t=wrap(q))<=pos && t<length; q=t) nr--;
    if(nr==0) return p;
    if(nr<0){
      do{p=wrap(p);}while(++nr); 
      return p;
      }
    pos=p-1;
    nr--;
    }
  return 0;
  }


// Backs up to the begin of the line preceding the line containing pos, or the 
// start of the line containing pos if the preceding line terminated in a newline
FXint FXText::changeBeg(FXint pos) const {
  register FXint p1,p2,t;
  FXASSERT(0<=pos && pos<=length);
  p1=p2=lineStart(pos);
  if(!(options&TEXT_WORDWRAP)) return p1;
  while(p2<pos && (t=wrap(p2))<=pos){
    p1=p2;
    p2=t;
    }
  return p1;
  }


// Scan forward to the end of affected area, which is the start of the next 
// paragraph; a change can cause the rest of the paragraph to reflow. 
FXint FXText::changeEnd(FXint pos) const {
  FXASSERT(0<=pos && pos<=length);
  while(pos<length){
    if(getChar(pos)=='\n') return pos+1;
    pos++;
    }
  return length+1;
  }



// Calculate line width
FXint FXText::lineWidth(FXint pos,FXint n) const {
  register FXint end=pos+n,w=0;
  FXASSERT(0<=pos && end<=length);
  while(pos<end){ w+=charWidth(getChar(pos),w); pos++; }
  return w;
  }


/*******************************************************************************/


// See if pos is a visible position
FXbool FXText::posVisible(FXint pos) const {
  return visrows[0]<=pos && pos<=visrows[nvisrows];
  }


// See if position is in the selection
FXbool FXText::isPosSelected(FXint pos) const {
  return hasSelection() && selstartpos<=pos && pos<=selendpos;
  }


// Find line number from visible pos
FXint FXText::posToLine(FXint pos,FXint ln) const {
  FXASSERT(0<=ln && ln<nvisrows);
  FXASSERT(visrows[ln]<=pos && pos<=visrows[nvisrows]);
  while(ln<nvisrows-1 && visrows[ln+1]<=pos && visrows[ln]<visrows[ln+1]) ln++;
  FXASSERT(0<=ln && ln<nvisrows);
  FXASSERT(visrows[ln]<=pos && pos<=visrows[ln+1]);
  return ln;
  }


// Localize position at x,y
FXint FXText::getPosAt(FXint x,FXint y) const {
  register FXint ln,ls,le,cx,cw;
  register FXchar ch;
  ln=(y-pos_y-margintop)/font->getFontHeight()-toprow;    // FIXME:- perhaps make this work outside visible area
  if(ln<0) return visrows[0];
  if(ln>=nvisrows) return visrows[nvisrows];
  x=x-pos_x-marginleft;
  if(x<0) return visrows[ln];
  ls=visrows[ln];
  le=visrows[ln+1];
  FXASSERT(0<=ls);
  FXASSERT(ls<=le);
  FXASSERT(le<=length);
  if(ls<le && isspace(getChar(le-1))) le--;      // Does NOT include last newline or blank
  cx=0;
  while(ls<le){
    ch=getChar(ls);
    cw=charWidth(ch,cx);
    if(x<=(cx+(cw>>1))) return ls;
    cx+=cw;
    ls+=1;
    }
  return le;
  }


// Determine x,y from position pos
FXint FXText::getYOfPos(FXint pos) const {
  register FXint h=font->getFontHeight();
  register FXint n,y;
  if(pos>length) pos=length;
  if(pos<0) pos=0;

  // Above visible part of buffer
  if(pos<visrows[0]){
    n=countRows(rowStart(pos),visrows[0]);
    y=(toprow-n)*h;
    FXTRACE((150,"getYOfPos(%d < visrows[0]=%d) = %d\n",pos,visrows[0],margintop+y));
    }

  // Below visible part of buffer
  else if(pos>visrows[nvisrows]){
    n=countRows(visrows[nvisrows-1],pos);
    y=(toprow+nvisrows-1+n)*h;
    FXTRACE((150,"getYOfPos(%d > visrows[%d]=%d) = %d\n",pos,nvisrows,visrows[nvisrows],margintop+y));
    }

  // In visible part of buffer
  else{
    n=posToLine(pos,0);
    y=(toprow+n)*h;
    FXTRACE((150,"getYOfPos(visrows[0]=%d <= %d <= visrows[%d]=%d) = %d\n",visrows[0],pos,nvisrows,visrows[nvisrows],margintop+y));
    }
  return margintop+y;
  }


// Calculate X position of pos
FXint FXText::getXOfPos(FXint pos) const {
  register FXint base=rowStart(pos);
  return marginleft+lineWidth(base,pos-base);
  }


// Force position to become fully visible
void FXText::makePositionVisible(FXint pos){
  FXint x,y,nx,ny;
  
  // Get coordinates of position
  x=getXOfPos(pos);
  y=getYOfPos(pos);

  // Check vertical visibility
  ny=pos_y;
  if(pos_y+y<margintop){
    ny=margintop-y;
    }
  else if(pos_y+y+font->getFontHeight()>viewport_h-marginbottom){
    ny=viewport_h-font->getFontHeight()-marginbottom-y;
    }

  // Check Horizontal visibility
  nx=pos_x;
  if(pos_x+x<marginleft){
    nx=marginleft-x;
    }
  else if(pos_x+x>viewport_w-marginright){
    nx=viewport_w-marginright-x;
    }

  // If needed, scroll
  if(nx!=pos_x || ny!=pos_y){
    setPosition(nx,ny);
    }
  }


// Make line containing pos the top visible line
void FXText::setTopLine(FXint pos){
  setPosition(pos_x,-getYOfPos(pos));
  }


// Make line containing pos the bottom visible line
void FXText::setBottomLine(FXint pos){
  setPosition(pos_x,viewport_h-font->getFontHeight()-marginbottom-getYOfPos(pos));
  }


// Center line of pos in the middle of the screen
void FXText::setCenterLine(FXint pos){
  setPosition(pos_x,viewport_h/2+font->getFontHeight()/2-getYOfPos(pos));
  }


// Get top line
FXint FXText::getTopLine() const { 
  return visrows[0]; 
  }


// Get bottom line
FXint FXText::getBottomLine() const { 
  return visrows[nvisrows-1]; 
  }


// Move content
void FXText::moveContents(FXint x,FXint y){
  register FXint delta,i,dx,dy,lineh;
  lineh=font->getFontHeight();

  // Erase fragments of cursor overhanging margins
  eraseCursorOverhang();
  
  // Number of lines scrolled
  delta=-y/lineh - toprow;

  // Scrolled up one or more lines
  if(delta<0){
    if(toprow+delta<=0){
      toppos=0;
      toprow=0;
      }
    else{
      toppos=prevRow(toppos,-delta);
      toprow=toprow+delta;
      }
    if(-delta<nvisrows){
      for(i=nvisrows; i>=-delta; i--) visrows[i]=visrows[delta+i];
      calcVisRows(0,-delta);
      }
    else{
      calcVisRows(0,nvisrows);
      }
    }
  
  // Scrolled down one or more lines
  else if(delta>0){
    if(toprow+delta>=nrows-1){
      toppos=rowStart(length);
      toprow=nrows-1;
      }
    else{
      toppos=nextRow(toppos,delta);
      toprow=toprow+delta;
      }
    if(delta<nvisrows){
      for(i=0; i<=nvisrows-delta; i++) visrows[i]=visrows[delta+i];
      calcVisRows(nvisrows-delta,nvisrows);
      }
    else{
      calcVisRows(0,nvisrows);
      }
    }
  
  // This is now the new keep position
  keeppos=toppos;
  
  // Hopefully, all is still in range
  FXASSERT(0<=toprow && toprow<=nrows-1);
  FXASSERT(0<=toppos && toppos<=length);

  // Scroll the contents
  dx=x-pos_x;
  dy=y-pos_y;
  pos_x=x;
  pos_y=y;
  scroll(marginleft,margintop,viewport_w-marginleft-marginright,viewport_h-margintop-marginbottom,dx,dy);
  }


/*******************************************************************************/


// Recalculate line starts
void FXText::calcVisRows(FXint startline,FXint endline){
  register FXint line,pos;
  FXASSERT(nvisrows>0);
  if(startline<0) 
    startline=0; 
  else if(startline>nvisrows) 
    startline=nvisrows;
  if(endline<0) 
    endline=0; 
  else if(endline>nvisrows) 
    endline=nvisrows;
  if(startline<=endline){
    FXASSERT(0<=toppos && toppos<=length);
    if(startline==0){
      visrows[0]=toppos;
      startline=1;
      }
    pos=visrows[startline-1];
    line=startline;
    if(options&TEXT_WORDWRAP){
      while(line<=endline && pos<length){
        pos=wrap(pos);
        visrows[line++]=pos;
        }
      }
    else{
      while(line<=endline && pos<length){
        pos=nextLine(pos);
        visrows[line++]=pos;
        }
      }
    while(line<=endline){
      visrows[line++]=length;
      }
    }
  }


// There has been a mutation in the buffer
void FXText::mutation(FXint pos,FXint ncins,FXint ncdel,FXint nrins,FXint nrdel){
  register FXint ncdelta=ncins-ncdel;
  register FXint nrdelta=nrins-nrdel;  
  register FXint line,i,x,y;

  FXTRACE((150,"BEFORE: pos=%d ncins=%d ncdel=%d nrins=%d nrdel=%d toppos=%d toprow=%d nrows=%d nvisrows=%d\n",pos,ncins,ncdel,nrins,nrdel,toppos,toprow,nrows,nvisrows));

  // All of the change is below the last visible line
  if(visrows[nvisrows]<pos){
    FXTRACE((150,"change below visible\n"));
    nrows+=nrdelta;
    }
  
  // All change above first visible line
  else if(pos+ncdel<=visrows[0]){
    FXTRACE((150,"change above visible\n"));
    nrows+=nrdelta;
    toprow+=nrdelta;
    toppos+=ncdelta;
    keeppos=toppos;
    for(i=0; i<=nvisrows; i++) visrows[i]+=ncdelta;
    pos_y-=nrdelta*font->getFontHeight();
    FXASSERT(0<=toppos && toppos<=length);
    }
  
  // Top visible part unchanged
  else if(visrows[0]<=pos){
    line=posToLine(pos,0);
    FXTRACE((150,"change below visible line %d\n",line));
    
    // More lines means paint the bottom half
    if(nrdelta>0){
      FXTRACE((150,"inserted %d rows\n",nrdelta));
      nrows+=nrdelta;
      for(i=nvisrows; i>line+nrdelta; i--) visrows[i]=visrows[i-nrdelta]+ncdelta;
      calcVisRows(line+1,line+nrins);
      FXASSERT(0<=toppos && toppos<=length);
      y=pos_y+margintop+(toprow+line)*font->getFontHeight();
      update(0,y,width,height-y);
      }

    // Less lines means paint bottom half also
    else if(nrdelta<0){
      FXTRACE((150,"deleted %d rows\n",-nrdelta));
      nrows+=nrdelta;
      for(i=line+1; i<=nvisrows+nrdelta; i++) visrows[i]=visrows[i-nrdelta]+ncdelta;
      calcVisRows(nvisrows+nrdelta,nvisrows);
      calcVisRows(line+1,line+nrins);
      FXASSERT(0<=toppos && toppos<=length);
      y=pos_y+margintop+(toprow+line)*font->getFontHeight();
      update(0,y,width,height-y);
      }

    // Same lines means paint the changed area only
    else{
      FXTRACE((150,"same number of rows\n"));
      for(i=line+1; i<=nvisrows; i++) visrows[i]=visrows[i]+ncdelta;
      calcVisRows(line+1,line+nrins);
      FXASSERT(0<=toppos && toppos<=length);
      if(nrins==0){
        x=pos_x+marginleft+lineWidth(visrows[line],pos-visrows[line]);
        y=pos_y+margintop+(toprow+line)*font->getFontHeight();
        update(x,y,width-x,font->getFontHeight());
        FXTRACE((150,"update(%d,%d,%d,%d)\n",x,y,width-x,font->getFontHeight()));
        }
      else{
        y=pos_y+margintop+(toprow+line)*font->getFontHeight();
        update(0,y,width,nrins*font->getFontHeight());
        FXTRACE((150,"update(%d,%d,%d,%d)\n",0,y,width,nrins*font->getFontHeight()));
        }
      }
    }
  
  // Bottom visible part unchanged
  else if(pos+ncdel<visrows[nvisrows-1]){
    nrows+=nrdelta;
    line=1+posToLine(pos+ncdel,0);
    FXASSERT(0<=line && line<nvisrows);
    FXASSERT(pos+ncdel<=visrows[line]);
    FXTRACE((150,"change above visible line %d\n",line));
    
    // Too few lines left to display
    if(toprow+nrdelta<=line){
      FXTRACE((150,"reset to top\n"));
      toprow=0;
      toppos=0;
      keeppos=0;
      pos_y=0;
      calcVisRows(0,nvisrows);
      FXASSERT(0<=toppos && toppos<=length);
      update();
      }
    
    // Redisplay only the top
    else{
      FXTRACE((150,"redraw top %d lines\n",line));
      toprow+=nrdelta;
      toppos=prevRow(visrows[line]+ncdelta,line);
      keeppos=toppos;
      pos_y-=nrdelta*font->getFontHeight();
      calcVisRows(0,nvisrows);
      FXASSERT(0<=toppos && toppos<=length);
      update(0,0,width,pos_y+margintop+(toprow+line)*font->getFontHeight());
      }
    }
  
  // All visible text changed
  else{
    FXTRACE((150,"change all visible lines\n"));
    nrows+=nrdelta;
    
    // Reset to top because too few lines left
    if(toprow>=nrows){
      FXTRACE((150,"reset to top\n"));
      toprow=0;
      toppos=0;
      keeppos=0;
      FXASSERT(0<=toppos && toppos<=length);
      pos_y=0;
      }
    
    // Maintain same row as before
    else{
      FXTRACE((150,"set to same row %d\n",toprow));
      toppos=nextRow(0,toprow);
      keeppos=toppos;
      FXASSERT(0<=toppos && toppos<=length);
      }
    calcVisRows(0,nvisrows);
    update();
    }
   
  FXTRACE((150,"AFTER : pos=%d ncins=%d ncdel=%d nrins=%d nrdel=%d toppos=%d toprow=%d nrows=%d\n",pos,ncins,ncdel,nrins,nrdel,toppos,toprow,nrows));
  }


// Replace m characters at pos by n characters
void FXText::replaceText(FXint pos,FXint m,const FXchar *text,FXint n){
  register FXint nrdel,nrins,ncdel,ncins,wbeg,wend,cursorstart;
  FXint wdel,hdel,wins,hins;
  if(n<0 || m<0 || pos<0 || length<pos+m){ fxerror("%s::replaceText: bad argument range.\n",getClassName()); }
  showCursor(0);    // FIXME can we do without this?
  
  FXTRACE((150,"pos=%d mdel=%d nins=%d\n",pos,m,n));
  
  // Bracket potentially affected character range for wrapping purposes
  wbeg=changeBeg(pos);
  wend=changeEnd(pos+m);
  
  // Measure stuff prior to change
  nrdel=measureText(wbeg,wend,wdel,hdel);
  ncdel=wend-wbeg;
    
  FXTRACE((150,"wbeg=%d wend=%d nrdel=%d ncdel=%d length=%d wdel=%d hdel=%d\n",wbeg,wend,nrdel,ncdel,length,wdel,hdel));
  
  // Modify the buffer
  sizegap(n-m);
  movegap(pos);
  memcpy(&buffer[pos],text,n);
  gapstart+=n;
  gapend+=m;
  length+=n-m;
  
  // Measure stuff after change
  nrins=measureText(wbeg,wend+n-m,wins,hins);
  ncins=wend+n-m-wbeg;

  FXTRACE((150,"wbeg=%d wend+n-m=%d nrins=%d ncins=%d length=%d wins=%d hins=%d\n",wbeg,wend+n-m,nrins,ncins,length,wins,hins));

  // Update stuff
  mutation(wbeg,ncins,ncdel,nrins,nrdel);

  // Fix text metrics
  textHeight=textHeight+hins-hdel;
  textWidth=FXMAX(textWidth,wins);

  // Reconcile scrollbars
  FXScrollArea::layout();     // FIXME:- scrollbars, but no layout

  // Fix selection end
  if(pos+m<=selendpos) selendpos+=n-m;
  else if(pos<=selendpos) selendpos=pos+n;

  // Fix selection start
  if(pos+m<=selstartpos) selstartpos+=n-m;
  else if(pos<=selstartpos) selstartpos=pos+n;

  // Fix anchor position
  if(pos+m<=anchorpos) anchorpos+=n-m;
  else if(pos<=anchorpos) anchorpos=pos+n;

  // Update cursor position variables
  if(wend<=cursorpos){
    cursorpos+=n-m;
    cursorrow+=nrins-nrdel;
    }
  else if(wbeg<=cursorpos){
    if(pos+m<=cursorpos) cursorpos+=n-m;
    else if(pos<=cursorpos) cursorpos=pos+n;
    cursorstart=rowStart(cursorpos);
    cursorcol=cursorpos-cursorstart;
    if(cursorstart<toppos)
      cursorrow=toprow-countRows(cursorstart,toppos);
    else
      cursorrow=toprow+countRows(toppos,cursorstart);
    }

  // Forget preferred column
  prefcol=-1;
  }


// Add text at the end
void FXText::appendText(const FXchar *text,FXint n){
  replaceText(length,0,text,n);
  }


// Insert some text at pos
void FXText::insertText(FXint pos,const FXchar *text,FXint n){
  replaceText(pos,0,text,n);
  }
    

// Remove some text at pos
void FXText::removeText(FXint pos,FXint n){
  replaceText(pos,n,NULL,0);
  }


// Grab range of text
void FXText::extractText(FXchar *text,FXint pos,FXint n) const {
  if(n<1 || pos<0 || length<pos+n){ fxerror("%s::extractText: bad argument range.\n",getClassName()); }
  if(!text){ fxerror("%s::extractText: NULL text argument.\n",getClassName()); }
  FXASSERT(0<n && 0<=pos && pos+n<=length);
  if(pos+n<=gapstart){
    memcpy(text,&buffer[pos],n);
    }
  else if(pos>=gapstart){
    memcpy(text,&buffer[pos-gapstart+gapend],n);
    }
  else{
    memcpy(text,&buffer[pos],gapstart-pos);
    memcpy(&text[gapstart-pos],&buffer[gapend],pos+n-gapstart);
    }
  }


// Completely reflow the text, because font, wrapwidth, or all of the
// text may have changed and everything needs to be recomputed
void FXText::recompute(){
  FXint hh=font->getFontHeight();
  FXint ww1,ww2,ww3,hh1,hh2,hh3,cursorstart;
  
  // Major recalc
  if(flags&FLAG_RECALC){

    // Make it point somewhere sensible
    if(keeppos<0) keeppos=0;
    if(keeppos>length) keeppos=length;

    // Make sure we're pointing to the start of a row again
    toppos=rowStart(keeppos);

    // Get start
    cursorstart=rowStart(cursorpos);
    cursorcol=cursorpos-cursorstart;
    
    // Avoid measuring huge chunks of text twice!
    if(cursorstart<toprow){
      cursorrow=measureText(0,cursorstart,ww1,hh1);
      toprow=cursorrow+measureText(cursorstart,toppos,ww2,hh2);
      nrows=toprow+measureText(toppos,length+1,ww3,hh3);
      }
    else{
      toprow=measureText(0,toppos,ww1,hh1);
      cursorrow=toprow+measureText(toppos,cursorstart,ww2,hh2);
      nrows=cursorrow+measureText(cursorstart,length+1,ww3,hh3);
      }
    
    textWidth=FXMAX3(ww1,ww2,ww3);
    textHeight=hh1+hh2+hh3;

    // Adjust position; we keep the same fractional position
    pos_y=-toprow*hh-(-pos_y%hh);
    }
  
  // Number of visible lines may have changed
  nvisrows=(getViewportHeight()+2*hh-2)/hh;
  
  // Number of visible lines changed; lines is 1 longer than nvisrows, 
  // so we can find the end of a line faster for every visible line
  FXRESIZE(&visrows,FXint,nvisrows+1);
  
  // Recompute line starts
  calcVisRows(0,nvisrows);
  
  FXTRACE((150,"recompute : toprow=%d toppos=%d nrows=%d nvisrows=%d textWidth=%d textHeight=%d length=%d cursorrow=%d cursorcol=%d\n",toprow,toppos,nrows,nvisrows,textWidth,textHeight,length,cursorrow,cursorcol));
  
  // Done with that
  flags&=~(FLAG_RECALC|FLAG_DIRTY);
  }


// Change the text in the buffer to new text
void FXText::setText(const FXchar* text,FXint n){
  if(n<0){ fxerror("%s::setText: bad argument range.\n",getClassName()); }
  if(!FXRESIZE(&buffer,FXchar,n+MINSIZE)){ fxerror("%s::setText: out of memory.\n",getClassName()); }
  memcpy(buffer,text,n);
  gapstart=n;
  gapend=gapstart+MINSIZE;
  length=n;
  toppos=0;
  toprow=0;
  keeppos=0;
  selstartpos=0;
  selendpos=0;
  anchorpos=0;
  cursorpos=0;
  cursorrow=0;
  cursorcol=0;
  prefcol=-1;
  pos_x=0;
  pos_y=0;
  recalc();
  layout();
  update();
  }


// Retrieve text into buffer
void FXText::getText(FXchar* text,FXint n) const {
  extractText(text,0,n);
  }


// Change all of the text
void FXText::setText(const FXString& text){
  setText(text.text(),text.length());
  }


// We return a constant copy of the buffer
FXString FXText::getText() const {
  FXString value;
  FXASSERT(0<=gapstart && gapstart<=length);
  value.append(buffer,gapstart);
  value.append(&buffer[gapend],length-gapstart);
  return value;
  }


// Change the font
void FXText::setFont(FXFont* fnt){
  if(!fnt){ fxerror("%s::setFont: NULL font specified.\n",getClassName()); }
  if(font!=fnt){
    font=fnt;
    recalc();
    tabwidth=tabcolumns*font->getTextWidth(" ",1);
    //if(options&TEXT_FIXEDWRAP){wrapwidth=wrapcolumns*font->getTextWidth(" ",1);}
    layout();
    update();
    }
  }


/*******************************************************************************/


// Determine content width of scroll area
FXint FXText::getContentWidth(){ 
  if(flags&FLAG_DIRTY) recompute();
  return marginleft+marginright+textWidth; 
  }


// Determine content height of scroll area
FXint FXText::getContentHeight(){ 
  if(flags&FLAG_DIRTY) recompute();
  return margintop+marginbottom+textHeight; 
  }


// Recalculate layout
void FXText::layout(){

  // Compute new wrap width
  if(!(options&TEXT_FIXEDWRAP)){
    wrapwidth=width-marginleft-marginright;
    if(!(options&VSCROLLER_NEVER)) wrapwidth-=vertical->getDefaultWidth();
    }
  else{
    wrapwidth=wrapcolumns*font->getTextWidth(" ",1);
    }
    
  // Scrollbars adjusted
  FXScrollArea::layout();
  
  // Set line size based on font
  vertical->setLine(font->getFontHeight());
  horizontal->setLine(font->getTextWidth(" ",1));
  
  // Force repaint
  update();
  
  // Done
  flags&=~FLAG_DIRTY;
  }


// The widget is resized
void FXText::resize(FXint w,FXint h){
  FXint hh=font->getFontHeight();
  FXint nv=(getViewportHeight()+2*hh-2)/hh;

  // In wrap mode, a width change causes a content recalculation
  if((options&TEXT_WORDWRAP) && !(options&TEXT_FIXEDWRAP) && (width!=w)) flags|=(FLAG_RECALC|FLAG_DIRTY);
  
  // Need to redo line starts
  if(nv!=nvisrows) flags|=FLAG_DIRTY;
  
  // Resize the window, and do layout
  FXScrollArea::resize(w,h);
  }


// The widget is moved and possibly resized
void FXText::position(FXint x,FXint y,FXint w,FXint h){
  FXint hh=font->getFontHeight();
  FXint nv=(getViewportHeight()+2*hh-2)/hh;
  
  // In wrap mode, a width change causes a content recalculation
  if((options&TEXT_WORDWRAP) && !(options&TEXT_FIXEDWRAP) && (width!=w)) flags|=(FLAG_RECALC|FLAG_DIRTY);

  // Need to redo line starts
  if(nv!=nvisrows) flags|=FLAG_DIRTY;

  // Place the window, and do layout
  FXScrollArea::position(x,y,w,h);
  }


/*******************************************************************************/


// Gained focus
long FXText::onFocusIn(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onFocusIn(sender,sel,ptr);
  if(!blinker){blinker=getApp()->addTimeout(getApp()->blinkSpeed,this,FXWindow::ID_CARETBLINK);}
  showCursor(FLAG_CARET);
  return 1;
  }


// Lost focus
long FXText::onFocusOut(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onFocusOut(sender,sel,ptr);
  if(blinker){blinker=getApp()->removeTimeout(blinker);}
  showCursor(FLAG_CARET);
  flags|=FLAG_UPDATE;
  return 1;
  }




// We were asked about status text
long FXText::onQueryHelp(FXObject* sender,FXSelector,void*){
  if(!help.empty() && (flags&FLAG_HELP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&help);
    return 1;
    }
  return 0;
  }


// Blink the cursor
long FXText::onBlink(FXObject*,FXSelector,void*){
  showCursor(flags^FLAG_CARET);
  blinker=getApp()->addTimeout(getApp()->blinkSpeed,this,FXWindow::ID_CARETBLINK);
  return 0;
  }
 

// Pressed left button
long FXText::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  flags&=~FLAG_TIP;
  handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    flags|=FLAG_PRESSED;
    flags&=~FLAG_UPDATE;
    if(ev->click_count==1){
      pos=getPosAt(ev->win_x,ev->win_y);
      FXTRACE((150,"getPosAt(%d,%d) = %d getYOfPos(%d) = %d getXOfPos(%d)=%d\n",ev->win_x,ev->win_y,pos,pos,getYOfPos(pos),pos,getXOfPos(pos)));
      if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
      if(ev->state&SHIFTMASK){
        extendSelection(pos,SELECT_CHARS);
        }
      else{
        killSelection();
        setAnchorPos(pos);
        }
      dragmode=MOUSE_CHARS;
      }
    
    // Select word by double click
    else if(ev->click_count==2){
      setAnchorPos(cursorpos);
      extendSelection(cursorpos,SELECT_WORDS);
      dragmode=MOUSE_WORDS;
      }
    
    // Select line by triple click
    else{
      setAnchorPos(cursorpos);
      extendSelection(cursorpos,SELECT_LINES);
      dragmode=MOUSE_LINES;
      }
    return 1;
    }
  return 0;
  }


// Released left button
long FXText::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){ 
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    stopAutoScroll();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    dragmode=MOUSE_NONE;
    return 1;
    }
  return 0;
  }


// Pressed middle button
long FXText::onMiddleBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  flags&=~FLAG_TIP;
  handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONPRESS),ptr)) return 1;
    pos=getPosAt(ev->win_x,ev->win_y);
    
    // Move over
    if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
    
    // Start text drag
    if(isPosSelected(pos)){
      dragmode=MOUSE_TRYDRAG;
      flags|=FLAG_PRESSED;
      flags&=~FLAG_UPDATE;
      }
 
    // Place cursor
    else if(isEditable()){
      dragmode=MOUSE_PASTE;
      flags|=FLAG_PRESSED;
      flags&=~FLAG_UPDATE;
      }
    
    // Beep
    else{
      getApp()->beep();
      }
    return 1;
    }
  return 0;
  }


// Released middle button
long FXText::onMiddleBtnRelease(FXObject*,FXSelector,void* ptr){
  FXuchar *data; FXuint len;
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    stopAutoScroll();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONRELEASE),ptr)) return 1;

    // Drop text somewhere
    if(dragmode==MOUSE_DRAG){
      handle(this,MKUINT(0,SEL_ENDDRAG),ptr);
      }

    // Paste text from selection (X-Windows only)
    else if(dragmode==MOUSE_PASTE){
      if(getDNDData(FROM_SELECTION,stringType,data,len)){
        FXRESIZE(&data,FXchar,len+1); data[len]='\0';
        handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
        FXFREE(&data);
        }
      }
    dragmode=MOUSE_NONE;
    return 1;
    }
  return 0;
  }


// Pressed right button
long FXText::onRightBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  flags&=~FLAG_TIP;
  handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONPRESS),ptr)) return 1;
    flags&=~FLAG_UPDATE;
    flags|=FLAG_PRESSED;
    grabx=ev->win_x-pos_x;
    graby=ev->win_y-pos_y;
    dragmode=MOUSE_SCROLL;
    return 1;
    }
  return 0;
  }


// Released right button
long FXText::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    dragmode=MOUSE_NONE;
    return 1;
    }
  return 0;
  }


// The widget lost the grab for some reason
long FXText::onUngrabbed(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onUngrabbed(sender,sel,ptr);
  flags&=~FLAG_PRESSED;
  flags|=FLAG_UPDATE;
  stopAutoScroll();
  return 1;
  }


// Autoscroll timer fired 
long FXText::onAutoScroll(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  FXScrollArea::onAutoScroll(sender,sel,ptr);
  switch(dragmode){
    case MOUSE_CHARS:
      if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
        pos=getPosAt(ev->win_x,ev->win_y);
        extendSelection(pos,SELECT_CHARS);
        if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
        }
      return 1;
    case MOUSE_WORDS:
      if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
        pos=getPosAt(ev->win_x,ev->win_y);
        extendSelection(pos,SELECT_WORDS);
        if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
        }
      return 1;
    case MOUSE_LINES:
      if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
        pos=getPosAt(ev->win_x,ev->win_y);
        extendSelection(pos,SELECT_LINES);
        if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
        }
      return 1;
    }
  return 0;
  }


// Handle real or simulated mouse motion
long FXText::onMotion(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  if(flags&FLAG_PRESSED){
    switch(dragmode){
      case MOUSE_CHARS:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          extendSelection(pos,SELECT_CHARS);
          if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      case MOUSE_WORDS:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          extendSelection(pos,SELECT_WORDS);
          if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      case MOUSE_LINES:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          extendSelection(pos,SELECT_LINES);
          if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      case MOUSE_SCROLL:
        setPosition(ev->win_x-grabx,ev->win_y-graby);
        break;
      case MOUSE_DRAG:
        handle(this,MKUINT(0,SEL_DRAGGED),ptr);
        break;
      case MOUSE_TRYDRAG:
        if(ev->moved){
          dragmode=MOUSE_NONE;
          if(handle(this,MKUINT(0,SEL_BEGINDRAG),ptr)){
            dragmode=MOUSE_DRAG;
            }
          }
        break;
      case MOUSE_PASTE:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>getApp()->dragDelta)||(fxabs(ev->win_y-ev->click_y)>getApp()->dragDelta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      }
    return 1;
    }
  return 0;
  }


// Cursor location changed
long FXText::onChanged(FXObject*,FXSelector,void* ptr){
  makePositionVisible((FXint)(long)ptr);
  setCursorPos((FXint)(long)ptr);
  if(target) target->handle(this,MKUINT(message,SEL_CHANGED),ptr);
  return 1;
  }


// Selected text
long FXText::onSelected(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_SELECTED),ptr);
  }


// Deselected text
long FXText::onDeselected(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_DESELECTED),ptr);
  }


// Inserted text
long FXText::onInserted(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_INSERTED),ptr);
  }


// Deleted text
long FXText::onDeleted(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_DELETED),ptr);
  }


/*******************************************************************************/


// Start a drag operation
long FXText::onBeginDrag(FXObject* sender,FXSelector sel,void* ptr){
  if(FXScrollArea::onBeginDrag(sender,sel,ptr)) return 1;
  beginDrag(&textType,1);
  setDragCursor(getApp()->dontdropCursor);
  return 1;
  }


// End drag operation
long FXText::onEndDrag(FXObject* sender,FXSelector sel,void* ptr){
  if(FXScrollArea::onEndDrag(sender,sel,ptr)) return 1;
  endDrag((didAccept()!=DRAG_REJECT));
  setDragCursor(getApp()->textCursor);
  return 1;
  }


// Dragged stuff around
long FXText::onDragged(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  FXDragAction action;
  if(FXScrollArea::onDragged(sender,sel,ptr)) return 1;
  action=DRAG_COPY;
  if(isEditable()){
    if(isDropTarget()) action=DRAG_MOVE;
    if(event->state&CONTROLMASK) action=DRAG_COPY;
    if(event->state&SHIFTMASK) action=DRAG_MOVE;
    }
  handleDrag(event->root_x,event->root_y,action);
  if(didAccept()!=DRAG_REJECT){
    if(action==DRAG_MOVE)
      setDragCursor(getApp()->dndMoveCursor);
    else
      setDragCursor(getApp()->dndCopyCursor);
    }
  else{
    setDragCursor(getApp()->dontdropCursor);
    }
  return 1;
  }


// Handle drag-and-drop enter
long FXText::onDNDEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onDNDEnter(sender,sel,ptr);
  return 1;
  }


// Handle drag-and-drop leave
long FXText::onDNDLeave(FXObject* sender,FXSelector sel,void* ptr){
  stopAutoScroll();
  FXScrollArea::onDNDLeave(sender,sel,ptr);
  return 1;
  }


// Handle drag-and-drop motion
long FXText::onDNDMotion(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXDragAction action;
  FXint pos;
    
  // Scroll into view  
  if(startAutoScroll(ev->win_x,ev->win_y,TRUE)) return 1;

  // Handled elsewhere
  if(FXScrollArea::onDNDMotion(sender,sel,ptr)) return 1;


//   FXDragType *types; FXuint numtypes;
//   inquireDNDTypes(FROM_DRAGNDROP,types,numtypes);
//   for(int i=0; i<numtypes; i++){
//     FXTRACE((100,"type %d = %s\n",types[i],getApp()->getDragTypeName(types[i]).text()));
//     }
 
  // Correct drop type
  if(offeredDNDType(FROM_DRAGNDROP,textType)){

    // Is target editable?
    if(isEditable()){
      action=inquireDNDAction();

      // Check for legal DND action
      if(action==DRAG_COPY || action==DRAG_MOVE){

        // Get the suggested drop position
        pos=getPosAt(ev->win_x,ev->win_y);

        // Move cursor to new position
        if(cursorpos!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);

        // We don't accept a drop on the selection
        if(!isPosSelected(pos)){
          acceptDrop(DRAG_ACCEPT);
          }
        }
      }
    return 1;
    }

  // Didn't handle it here
  return 0;
  }


// Handle drag-and-drop drop
long FXText::onDNDDrop(FXObject* sender,FXSelector sel,void* ptr){
  FXuchar *data,*junk; FXuint len,dum;

  // Stop scrolling
  stopAutoScroll();

  // Try handling it in base class first
  if(FXScrollArea::onDNDDrop(sender,sel,ptr)) return 1;
  
  // Should really not have gotten this if non-editable
  if(isEditable()){
    
    // Try handle here
    if(getDNDData(FROM_DRAGNDROP,textType,data,len)){
      FXRESIZE(&data,FXchar,len+1); data[len]='\0';
 
      // Need to ask the source to delete his copy
      if(inquireDNDAction()==DRAG_MOVE){
        getDNDData(FROM_DRAGNDROP,deleteType,junk,dum);
        FXASSERT(!junk);
        }

      // Insert the new text
      handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
      FXFREE(&data);
      }
    return 1;
    }
  return 0;
  }


// Service requested DND data
long FXText::onDNDRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data; FXuint len;
  
  // Perhaps the target wants to supply its own data
  if(FXScrollArea::onDNDRequest(sender,sel,ptr)) return 1;
  
  // Return dragged text
  if(event->target==textType){
    len=selendpos-selstartpos;
    FXMALLOC(&data,FXuchar,len);
    extractText((FXchar*)data,selstartpos,len);
    setDNDData(FROM_DRAGNDROP,stringType,data,len);
    return 1;
    }

  // Delete dragged text
  if(event->target==deleteType){
    if(isEditable()){
      handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
      }
    return 1;
    }

  return 0;
  }


/*******************************************************************************/


// We now really do have the selection
long FXText::onSelectionGained(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onSelectionGained(sender,sel,ptr);
  return 1;
  }


// We lost the selection somehow 
long FXText::onSelectionLost(FXObject* sender,FXSelector sel,void* ptr){
  FXbool hadselection=hasSelection();
  FXScrollArea::onSelectionLost(sender,sel,ptr);
  if(hadselection) updateRange(selstartpos,selendpos);
  return 1;
  }


// Somebody wants our selection
long FXText::onSelectionRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data; FXuint len;

  // Perhaps the target wants to supply its own data for the selection
  if(FXScrollArea::onSelectionRequest(sender,sel,ptr)) return 1;
  
  // Return text of the selection
  if(event->target==stringType){
    FXASSERT(selstartpos<=selendpos);
    len=selendpos-selstartpos;
    FXMALLOC(&data,FXuchar,len);
    extractText((FXchar*)data,selstartpos,len);
    setDNDData(FROM_SELECTION,stringType,data,len);
    return 1;
    }

  return 0;
  }


/*******************************************************************************/
   
   
// We now really do have the selection
long FXText::onClipboardGained(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onClipboardGained(sender,sel,ptr);
  return 1;
  }


// We lost the selection somehow 
long FXText::onClipboardLost(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onClipboardLost(sender,sel,ptr);
  FXFREE(&clipbuffer);
  clipbuffer=NULL;
  cliplength=0;
  return 1;
  }


// Somebody wants our selection
long FXText::onClipboardRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data;

  // Try handling it in base class first
  if(FXScrollArea::onClipboardRequest(sender,sel,ptr)) return 1;

  // Requested data from clipboard
  if(event->target==stringType){
    FXASSERT(cliplength);
#ifndef WIN32
    FXMALLOC(&data,FXuchar,cliplength);
    memcpy(data,clipbuffer,cliplength);
    setDNDData(FROM_CLIPBOARD,stringType,data,cliplength);
#else
    FXint i,j,c;
    FXCALLOC(&data,FXuchar,2*cliplength+1);
    for(i=j=0; j<cliplength; j++){
      c=clipbuffer[j]; 
      if(c=='\n'){data[i++]='\r';data[i++]='\n';}else{data[i++]=c;}
      }
    setDNDData(FROM_CLIPBOARD,stringType,data,i+1);
#endif
    return 1;
    }

  return 0;
  }

/*******************************************************************************/


// Keyboard press
long FXText::onKeyPress(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  char entry;
  flags&=~FLAG_TIP;
  if(!isEnabled()) return 0;
  FXTRACE((200,"%s::onKeyPress keysym=0x%04x state=%04x\n",getClassName(),event->code,event->state));
  if(target && target->handle(this,MKUINT(message,SEL_KEYPRESS),ptr)) return 1;
  if(!(event->state&ALTMASK)){
    switch(event->code){
      case KEY_Shift_L:
      case KEY_Shift_R:
      case KEY_Control_L:
      case KEY_Control_R:
        if(dragmode==MOUSE_DRAG){
          handle(this,MKUINT(0,SEL_DRAGGED),ptr);
          }
        return 1;
      case KEY_Up:
      case KEY_KP_Up:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        handle(this,MKUINT(ID_CURSOR_UP,SEL_COMMAND),NULL);
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Down:
      case KEY_KP_Down:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        handle(this,MKUINT(ID_CURSOR_DOWN,SEL_COMMAND),NULL);
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Left:
      case KEY_KP_Left:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        if(event->state&CONTROLMASK){
          handle(this,MKUINT(ID_CURSOR_WORD_LEFT,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_CURSOR_LEFT,SEL_COMMAND),NULL);
          }
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Right:
      case KEY_KP_Right:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        if(event->state&CONTROLMASK){
          handle(this,MKUINT(ID_CURSOR_WORD_RIGHT,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_CURSOR_RIGHT,SEL_COMMAND),NULL);
          }
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Home:
      case KEY_KP_Home:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        if(event->state&CONTROLMASK){
          handle(this,MKUINT(ID_CURSOR_TOP,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_CURSOR_HOME,SEL_COMMAND),NULL);
          }
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_End:
      case KEY_KP_End:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        if(event->state&CONTROLMASK){
          handle(this,MKUINT(ID_CURSOR_BOTTOM,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_CURSOR_END,SEL_COMMAND),NULL);
          }
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Page_Up:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        handle(this,MKUINT(ID_CURSOR_PAGEUP,SEL_COMMAND),NULL);
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Page_Down:
        if(!(event->state&SHIFTMASK)){
          handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          }
        handle(this,MKUINT(ID_CURSOR_PAGEDOWN,SEL_COMMAND),NULL);
        if(event->state&SHIFTMASK){
          handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
          }
        return 1;
      case KEY_Insert:
      case KEY_KP_Insert:
        if(event->state&CONTROLMASK){
          handle(this,MKUINT(ID_COPY_SEL,SEL_COMMAND),NULL);
          }
        else if(event->state&SHIFTMASK){
          if(isEditable()){
            handle(this,MKUINT(ID_PASTE_SEL,SEL_COMMAND),NULL);
            }
          else{
            getApp()->beep();
            }
          }
        else{
          options^=TEXT_OVERSTRIKE;
          }
        return 1;
      case KEY_Delete:
      case KEY_KP_Delete:
        if(isEditable()){
          if(isPosSelected(cursorpos)){
            handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
            }
          else{
            handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
            if(event->state&CONTROLMASK){
              handle(this,MKUINT(ID_DELETE_WORD,SEL_COMMAND),NULL);
              }
            else if(event->state&SHIFTMASK){
              handle(this,MKUINT(ID_DELETE_EOL,SEL_COMMAND),NULL);
              }
            else{
              handle(this,MKUINT(ID_DELETE,SEL_COMMAND),NULL);
              }
            }
          }
        else{
          getApp()->beep();
          }
        return 1;
      case KEY_BackSpace:
        if(isEditable()){
          if(isPosSelected(cursorpos)){
            handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
            }
          else{
           handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
           if(event->state&CONTROLMASK){
              handle(this,MKUINT(ID_BACKSPACE_WORD,SEL_COMMAND),NULL);
              }
            else if(event->state&SHIFTMASK){
              handle(this,MKUINT(ID_BACKSPACE_BOL,SEL_COMMAND),NULL);
              }
            else{
              handle(this,MKUINT(ID_BACKSPACE,SEL_COMMAND),NULL);
              }
            }
          }
        else{
          getApp()->beep();
          }
        return 1;
      case KEY_Return:
      case KEY_KP_Enter:
      case KEY_ISO_Enter:
        if(isEditable()){
          if(isPosSelected(cursorpos)){
            handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
            }
          else{
            handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
            }
          handle(this,MKUINT(ID_INSERT_NEWLINE,SEL_COMMAND),NULL);
          }
        else{
          getApp()->beep();
          }
        return 1;
      default:
        if(event->state&CONTROLMASK){   // Control key
          switch(event->code){
            case KEY_a:
              handle(this,MKUINT(ID_SELECT_ALL,SEL_COMMAND),NULL);
              return 1;
            case KEY_x:
              if(isEditable()){
                handle(this,MKUINT(ID_CUT_SEL,SEL_COMMAND),NULL);
                }
              else{
                getApp()->beep();
                }
              return 1;
            case KEY_c:
              handle(this,MKUINT(ID_COPY_SEL,SEL_COMMAND),NULL);
              return 1;
            case KEY_v:
              if(isEditable()){
                handle(this,MKUINT(ID_PASTE_SEL,SEL_COMMAND),NULL);
                }
              else{
                getApp()->beep();
                }
              return 1;
            case KEY_k:
              if(isEditable()){
                handle(this,MKUINT(ID_DELETE_LINE,SEL_COMMAND),NULL);
                }
              else{
                getApp()->beep();
                }
              return 1;
            case KEY_h:
              handle(this,MKUINT(ID_CURSOR_SCRNTOP,SEL_COMMAND),NULL);
              return 1;
            case KEY_m:
              handle(this,MKUINT(ID_CURSOR_SCRNCTR,SEL_COMMAND),NULL);
              return 1;
            case KEY_l:
              handle(this,MKUINT(ID_CURSOR_SCRNBTM,SEL_COMMAND),NULL);
              return 1;
            }
          }
        else{                           // Normal key
          entry=fxkeyval(event->code,event->state);
          if((' '<=(FXuchar)entry)||(entry=='\t')){
            if(isEditable()){
              if(isPosSelected(cursorpos)){
                handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
                }
              else{
                handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
                }
              handle(this,MKUINT(ID_INSERT_CHAR,SEL_COMMAND),(void*)(long)entry);
              }
            else{
              getApp()->beep();
              }
            }
          return 1;
          }
        break;
      }
    }
  return 0;
  }
  
  
// Keyboard release
long FXText::onKeyRelease(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(!isEnabled()) return 0;
  FXTRACE((200,"%s::onKeyRelease keysym=0x%04x state=%04x\n",getClassName(),event->code,event->state));
  if(target && target->handle(this,MKUINT(message,SEL_KEYRELEASE),ptr)) return 1;
  if(!(event->state&ALTMASK)){
    switch(event->code){
      case KEY_Shift_L:
      case KEY_Shift_R:
      case KEY_Control_L:
      case KEY_Control_R:
        if(dragmode==MOUSE_DRAG){
          handle(this,MKUINT(0,SEL_DRAGGED),ptr);
          }
        return 1;
      case KEY_Up:
      case KEY_KP_Up:
      case KEY_Down:
      case KEY_KP_Down:
      case KEY_Left:
      case KEY_KP_Left:
      case KEY_Right:
      case KEY_KP_Right:
      case KEY_Home:
      case KEY_KP_Home:
      case KEY_End:
      case KEY_KP_End:
      case KEY_Page_Up:
      case KEY_Page_Down:
      case KEY_Insert:
      case KEY_KP_Insert:
      case KEY_Delete:
      case KEY_KP_Delete:
      case KEY_BackSpace:
      case KEY_Return:
      case KEY_KP_Enter:
      case KEY_ISO_Enter:
        return 1;
      default:
        if(event->state&CONTROLMASK){   // Control key
          switch(event->code){
            case KEY_a:
            case KEY_x:
            case KEY_c:
            case KEY_v:
            case KEY_k:
            case KEY_h:
            case KEY_m:
            case KEY_l:
              return 1;
            }
          }
        else{                           // Normal key
          return 1;
          }
        break;
      }
    }
  return 0;
  }


/*******************************************************************************/

// Move cursor to top of buffer
long FXText::onCmdCursorTop(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)0);
  }


// Move cursor to bottom of buffer
long FXText::onCmdCursorBottom(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)length);
  }


// Move cursor to begin of line
long FXText::onCmdCursorHome(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)rowStart(cursorpos));
  }


// Move cursor to end of line
long FXText::onCmdCursorEnd(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)rowEnd(cursorpos));
  }


// Move cursor right
long FXText::onCmdCursorRight(FXObject*,FXSelector,void*){
  if(cursorpos>=length) return 1;
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)(cursorpos+1));
  }


// Move cursor left
long FXText::onCmdCursorLeft(FXObject*,FXSelector,void*){
  if(cursorpos<=0) return 1;
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)(cursorpos-1));
  }


// Move cursor to previous line
long FXText::onCmdCursorUp(FXObject*,FXSelector,void*){
  FXint newrow,newpos,rowend,col;
  col=(0<=prefcol)?prefcol:cursorcol;
  newrow=prevRow(cursorpos);
  newpos=newrow+col;
  rowend=rowEnd(newrow);
  if(newpos>rowend) newpos=rowend;
  handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
  prefcol=col;
  return 1;
  }


// Move cursor to next line
long FXText::onCmdCursorDown(FXObject*,FXSelector,void*){
  FXint newrow,newpos,rowend,col;
  col=(0<=prefcol)?prefcol:cursorcol;
  newrow=nextRow(cursorpos);
  newpos=newrow+col;
  rowend=rowEnd(newrow);
  if(newpos>rowend) newpos=rowend;
  handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
  prefcol=col;
  return 1;
  }


// Page down
long FXText::onCmdCursorPageDown(FXObject*,FXSelector,void*){
  FXint newrow,newpos,rowend,col;
  col=(0<=prefcol)?prefcol:cursorcol;
  newrow=nextRow(cursorpos,(viewport_h)/font->getFontHeight());
  newpos=newrow+col;
  rowend=rowEnd(newrow);
  if(newpos>rowend) newpos=rowend;
  setTopLine(nextRow(toppos,viewport_h/font->getFontHeight()));
  //handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
  setCursorPos(newpos);
  prefcol=col;
  return 1;
  }


// Page up
long FXText::onCmdCursorPageUp(FXObject*,FXSelector,void*){
  FXint newrow,newpos,rowend,col;
  col=(0<=prefcol)?prefcol:cursorcol;
  newrow=prevRow(cursorpos,(viewport_h)/font->getFontHeight());
  newpos=newrow+col;
  rowend=rowEnd(newrow);
  if(newpos>rowend) newpos=rowend;
  setTopLine(prevRow(toppos,viewport_h/font->getFontHeight()));
  //handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
  setCursorPos(newpos);
  prefcol=col;
  return 1;
  }


// Word Left
long FXText::onCmdCursorWordLeft(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)leftWord(cursorpos));
  }


// Word Right
long FXText::onCmdCursorWordRight(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)rightWord(cursorpos));
  }


// Cursor pos to top of screen
long FXText::onCmdCursorScreenTop(FXObject*,FXSelector,void*){
  setTopLine(cursorpos);
  return 1;
  }


// Cursor pos to bottom of screen
long FXText::onCmdCursorScreenBottom(FXObject*,FXSelector,void*){
  setBottomLine(cursorpos);
  return 1;
  }


// Cursor pos to center of screen
long FXText::onCmdCursorScreenCenter(FXObject*,FXSelector,void*){
  setCenterLine(cursorpos);
  return 1;
  }


// Move cursor to begin of paragraph
long FXText::onCmdCursorParHome(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)lineStart(cursorpos));
  }


// Move cursor to end of paragraph
long FXText::onCmdCursorParEnd(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)lineEnd(cursorpos));
  }


// Mark
long FXText::onCmdMark(FXObject*,FXSelector,void*){
  setAnchorPos(cursorpos);
  return 1;
  }


// Extend
long FXText::onCmdExtend(FXObject*,FXSelector,void*){
  extendSelection(cursorpos);
  return 1;
  }


// Insert a character
long FXText::onCmdInsertChar(FXObject*,FXSelector,void* ptr){
  FXchar ch=(FXchar)(long)ptr;
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=cursorpos;
  what[1]=1;
  if(options&TEXT_OVERSTRIKE){    // FIXME overstriking is still BUGGY
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    replaceText(what[0],1,&ch,1);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  else{
    insertText(what[0],&ch,1);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Insert a string
long FXText::onCmdInsertString(FXObject*,FXSelector,void* ptr){
  FXchar* string=(FXchar*)ptr;
  FXint what[2];
  if(!isEditable()) return 1;
  FXASSERT(string);
  what[0]=cursorpos;
  what[1]=strlen(string);
  if(options&TEXT_OVERSTRIKE){    // FIXME overstriking is still BUGGY
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    replaceText(what[0],what[1],string,what[1]);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  else{
    insertText(what[0],string,what[1]);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Insert a character
long FXText::onCmdInsertNewline(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=cursorpos;
  what[1]=1;
  insertText(what[0],"\n",1);
  handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Cut
long FXText::onCmdCutSel(FXObject*,FXSelector,void*){
  FXDragType types[1];
  if(hasSelection()){
    if(isEditable()){
      types[0]=stringType;
      if(acquireClipboard(types,1)){
        FXFREE(&clipbuffer);
        FXASSERT(selstartpos<=selendpos);
        cliplength=selendpos-selstartpos;
        FXCALLOC(&clipbuffer,FXchar,cliplength+1);
        if(!clipbuffer){ 
          fxwarning("%s::onCmdCutSel: out of memory\n",getClassName()); 
          cliplength=0;
          }
        else{
          extractText(clipbuffer,selstartpos,cliplength);
          handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
          }
        }
      }
    else{
      getApp()->beep();
      }
    }
  return 1;
  }


// Copy
long FXText::onCmdCopySel(FXObject*,FXSelector,void*){
  FXDragType types[1];
  if(hasSelection()){
    types[0]=stringType;
    if(acquireClipboard(types,1)){
      FXFREE(&clipbuffer);
      FXASSERT(selstartpos<=selendpos);
      cliplength=selendpos-selstartpos;
      FXCALLOC(&clipbuffer,FXchar,cliplength+1);
      if(!clipbuffer){ 
        fxwarning("%s::onCmdCopySel: out of memory\n",getClassName()); 
        cliplength=0;
        }
      else{
        extractText(clipbuffer,selstartpos,cliplength);
        }
      }
    }
  return 1;
  }


// Paste
long FXText::onCmdPasteSel(FXObject*,FXSelector,void*){
  FXuchar *data; FXuint len;
  if(isEditable()){
    if(hasSelection()){
      handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
      }
    if(getDNDData(FROM_CLIPBOARD,stringType,data,len)){
      FXRESIZE(&data,FXuchar,len+1); data[len]='\0';
#ifdef WIN32
      FXuint i,j;
      FXuchar c;
      for(i=j=0; j<len; j++){if((c=data[j])!='\r'){data[i++]=c;}}
      data[i]='\0';
#endif
      handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
      FXFREE(&data);
      return 1;
      }
    }
  else{
    getApp()->beep();
    }
  return 1;
  }


// Delete selection
long FXText::onCmdDeleteSel(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!hasSelection()) return 1;
  what[0]=selstartpos;
  what[1]=selendpos-selstartpos;
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  killSelection();
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Select character
long FXText::onCmdSelectChar(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(cursorpos);
  extendSelection(cursorpos+1,SELECT_CHARS);
  what[0]=selstartpos;
  what[1]=1;
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Select Word
long FXText::onCmdSelectWord(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(cursorpos);
  extendSelection(cursorpos,SELECT_WORDS);
  what[0]=selstartpos;
  what[1]=selendpos-what[0];
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Select Line
long FXText::onCmdSelectLine(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(cursorpos);
  extendSelection(cursorpos,SELECT_LINES);
  what[0]=selstartpos;
  what[1]=selendpos-what[0];
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Select All
long FXText::onCmdSelectAll(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(0);
  extendSelection(length,SELECT_CHARS);
  what[0]=selstartpos;
  what[1]=selendpos-what[0];
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Deselect All
long FXText::onCmdDeselectAll(FXObject*,FXSelector,void*){
  FXint what[2];
  what[0]=selstartpos;
  what[1]=selendpos-what[0];
  handle(this,MKUINT(0,SEL_DESELECTED),(void*)what);
  killSelection();
  return 1;
  }


// Backspace character
long FXText::onCmdBackspace(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  if(cursorpos==0){ getApp()->beep(); return 1; }
  what[0]=cursorpos-1;
  what[1]=1;
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Backspace word
long FXText::onCmdBackspaceWord(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=leftWord(cursorpos);
  what[1]=cursorpos-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Backspace bol
long FXText::onCmdBackspaceBol(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=rowStart(cursorpos);
  what[1]=cursorpos-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Delete character
long FXText::onCmdDelete(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  if(cursorpos==length){ getApp()->beep(); return 1; }
  what[0]=cursorpos;
  what[1]=1;
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Delete word
long FXText::onCmdDeleteWord(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=cursorpos;
  what[1]=rightWord(cursorpos)-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Delete end of line
long FXText::onCmdDeleteEol(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=cursorpos;
  what[1]=rowEnd(cursorpos)-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }


// Delete line
long FXText::onCmdDeleteLine(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=rowStart(cursorpos);
  what[1]=nextRow(cursorpos)-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)cursorpos);
  flags|=FLAG_CHANGED;
  modified=TRUE;
  return 1;
  }

/*******************************************************************************/

// Editable toggle
long FXText::onCmdToggleEditable(FXObject*,FXSelector,void*){
  options^=TEXT_READONLY;
  return 1;
  }


// Update editable toggle
long FXText::onUpdToggleEditable(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&TEXT_READONLY) ? ID_UNCHECK : ID_CHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }


// Overstrike toggle
long FXText::onCmdToggleOverstrike(FXObject*,FXSelector,void*){
  options^=TEXT_OVERSTRIKE;
  return 1;
  }


// Update overstrike toggle
long FXText::onUpdToggleOverstrike(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&TEXT_OVERSTRIKE) ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }


// Move cursor to indicated row
long FXText::onCmdCursorRow(FXObject* sender,FXSelector,void*){
  FXint row,col,newrow,newpos,rowend;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&row)){
    if(row<1) row=1;
    if(row>nrows) row=nrows;
    --row;
    if(row!=cursorrow){
      col=(0<=prefcol)?prefcol:cursorcol;
      if(row>cursorrow)
        newrow=nextRow(cursorpos,row-cursorrow);
      else
        newrow=prevRow(cursorpos,cursorrow-row);
      newpos=newrow+col;
      rowend=rowEnd(newrow);
      if(newpos>rowend) newpos=rowend;
      handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
      prefcol=col;
      }
    }
  return 1;
  }


// Being asked about current row number
long FXText::onUpdCursorRow(FXObject* sender,FXSelector,void*){
  FXint row=cursorrow+1;
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&row);
  return 1;
  }


// Move cursor to indicated column
long FXText::onCmdCursorColumn(FXObject* sender,FXSelector,void*){
  FXint col,newpos,rowbeg,rowend;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&col)){
    rowbeg=cursorpos-cursorcol;
    rowend=rowEnd(cursorpos);
    newpos=rowbeg+col;
    if(newpos<rowbeg) newpos=rowbeg;
    if(newpos>rowend) newpos=rowend;
    handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
    }
  return 1;
  }


// Being asked about current column
long FXText::onUpdCursorColumn(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&cursorcol);
  return 1;
  }


// Update somebody who works on the selection
long FXText::onUpdHaveSelection(FXObject* sender,FXSelector,void* ptr){
  FXuint msg=hasSelection() ? ID_ENABLE : ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Tab columns
long FXText::onUpdTabColumns(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&tabcolumns);
  return 1;
  }


// Change tab columns
long FXText::onCmdTabColumns(FXObject* sender,FXSelector,void*){
  FXint cols;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&cols)){
    setTabColumns(cols);
    }
  return 1;
  }


// Wrap columns
long FXText::onUpdWrapColumns(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&wrapcolumns);
  return 1;
  }


// Change wrap columns
long FXText::onCmdWrapColumns(FXObject* sender,FXSelector,void*){
  FXint cols;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&cols)){
    setWrapColumns(cols);
    }
  return 1;
  }

/*******************************************************************************/


// Obtain text style at position pos; note column may be outside of text
// to allow for rectangular selections!
FXuint FXText::getStyleAt(FXint linepos,FXint linelen,FXint column) const {
  register FXuint style=0;
  register FXint pos;
  register FXchar ch;

  // Line pos is legal?
  FXASSERT(0<=linepos && linepos<=length);

  // Get legal position
  pos=linepos+FXMIN(column,linelen);

  // Position is legal
  FXASSERT(0<=pos && pos<=length);

  // Blank part of line
  if(column>=linelen) style|=STYLE_FILL;

  // Selected part of text
  if(hasSelection() && selstartpos<=pos && pos<selendpos) style|=STYLE_SELECTED;

  // There's text here
  if(pos<linepos+linelen){

    // Special style for control characters
    ch=getChar(pos);

    // Get special style
    if(ch=='\t') style|=STYLE_FILL; 
    else if(((FXuchar)ch)<' ') style|=STYLE_CONTROL; 
    }
  return style;
  }


// Draw chunk from buffer
void FXText::drawBufString(FXDCWindow& dc,FXint x,FXint y,FXint w,FXint h,FXint pos,FXint n,FXuint style){
  FXchar ch;

  // Filled space 
  if(style&STYLE_FILL){
    if(style&STYLE_SELECTED){
      dc.setForeground(selbackColor);
      }
    else{
      dc.setForeground(backColor);
      }
    dc.fillRectangle(x,y,w,h);
    }

  // String of control characters
  else if(style&STYLE_CONTROL){
    FXASSERT(0<=pos && pos+n<=length);
    y+=font->getFontAscent();
    if(style&STYLE_SELECTED){
      dc.setForeground(seltextColor);
      dc.setBackground(selbackColor);
      }
    else{
      dc.setForeground(textColor);
      dc.setBackground(backColor);
      }
    while(pos<gapstart && 0<n){
      ch=buffer[pos]|0x40;
      dc.drawImageText(x,y,"^",1);
      x+=font->getTextWidth("^",1);
      dc.drawImageText(x,y,&ch,1);
      x+=font->getTextWidth(&ch,1);
      pos++;
      n--;
      }
    while(0<n){
      ch=buffer[pos-gapstart+gapend]|0x40;
      dc.drawImageText(x,y,"^",1);
      x+=font->getTextWidth("^",1);
      dc.drawImageText(x,y,&ch,1);
      x+=font->getTextWidth(&ch,1);
      pos++;
      n--;
      }
    }

  // Regular text
  else{
    FXASSERT(0<=pos && pos+n<=length);
    y+=font->getFontAscent();
    if(style&STYLE_SELECTED){
      dc.setForeground(seltextColor);
      dc.setBackground(selbackColor);
      }
    else{
      dc.setForeground(textColor);
      dc.setBackground(backColor);
      }
    if(pos+n<=gapstart){
      dc.drawImageText(x,y,&buffer[pos],n);
      }
    else if(pos>=gapstart){
      dc.drawImageText(x,y,&buffer[pos-gapstart+gapend],n);
      }
    else{
      dc.drawImageText(x,y,&buffer[pos],gapstart-pos);
      x+=font->getTextWidth(&buffer[pos],gapstart-pos);
      dc.drawImageText(x,y,&buffer[gapend],pos+n-gapstart);
      }
    if(style&STYLE_UNDERLINE) dc.drawLine(x,y+1,x+w,y+1);
//     if(style&STYLE_SELECTED){
//       dc.setForeground(selbackColor);
//       dc.fillRectangle(x,y,w,h);
//       dc.setForeground(seltextColor);
//       }
//     else{
//       dc.setForeground(backColor);
//       dc.fillRectangle(x,y,w,h);
//       dc.setForeground(textColor);
//       }
//     y+=font->getFontAscent();
//     if(pos+n<=gapstart){
//       dc.drawText(x,y,&buffer[pos],n);
//       }
//     else if(pos>=gapstart){
//       dc.drawText(x,y,&buffer[pos-gapstart+gapend],n);
//       }
//     else{
//       dc.drawText(x,y,&buffer[pos],gapstart-pos);
//       x+=font->getTextWidth(&buffer[pos],gapstart-pos);
//       dc.drawText(x,y,&buffer[gapend],pos+n-gapstart);
//       }
//     if(style&STYLE_UNDERLINE) dc.drawLine(x,y+1,x+w,y+1);
    }
  }


// Draw partial text line with correct style
void FXText::drawTextRow(FXDCWindow& dc,FXint line,FXint lclip,FXint rclip,FXint fc,FXint tc){
  register FXint x,y,w,h,linebeg,lineend,linelen,sc,ec,cw,sw;
  register FXuint curstyle,newstyle;
  register FXchar c;
  FXASSERT(0<=fc && fc<=tc);
  FXASSERT(lclip<rclip);
  FXASSERT(0<=line && line<=nvisrows);
  linebeg=visrows[line];
  lineend=visrows[line+1];
  if(linebeg<lineend && isspace(getChar(lineend-1))) lineend--;
  linelen=lineend-linebeg;
  w=0;
  h=font->getFontHeight();
  x=0;
  y=pos_y+margintop+(toprow+line)*h;
  sw=font->getTextWidth(" ",1);
  for(sc=0; ; sc++){
    if(sc>=linelen){
      cw=sw;                  // For selections outside of text 
      }
    else{
      c=getChar(linebeg+sc);
      cw=charWidth(c,x);
      }
    if(sc>=fc && x+pos_x+marginleft+cw>=lclip) break;
    x+=cw;
    }
  FXASSERT(fc<=sc);
  curstyle=newstyle=getStyleAt(linebeg,linelen,sc);
  for(ec=sc; ec<=tc; ec++){
    newstyle=getStyleAt(linebeg,linelen,ec);
    if(newstyle!=curstyle){
      drawBufString(dc,pos_x+marginleft+x,y,w,h,linebeg+sc,ec-sc,curstyle);
      curstyle=newstyle;
      x+=w;
      w=0;
      sc=ec;
      }
    if(ec>=linelen){
      cw=sw;                    // For selections outside of text 
      }
    else{
      c=getChar(linebeg+ec);
      cw=charWidth(c,x+w);
      }
    if(x+pos_x+marginleft+w>=rclip) break;
    w+=cw;
    }
  drawBufString(dc,pos_x+marginleft+x,y,w,h,linebeg+sc,ec-sc,curstyle);
  }


// Show or hide cursor
void FXText::showCursor(FXuint state){
  if((state^flags)&FLAG_CARET){
    flags^=FLAG_CARET;
    drawCursor(flags&FLAG_CARET);
    }
  }


// Draw the cursor
void FXText::drawCursor(FXuint state){
  register FXint xx,yt,yb,fh;
  if(!id()) return;
  if(visrows[0]<=cursorpos && cursorpos<=visrows[nvisrows]){
    FXASSERT(0<=cursorpos && cursorpos<=length);
    FXASSERT(0<=cursorrow && cursorrow<nrows);
    FXASSERT(0<=(cursorrow-toprow));
    FXASSERT((cursorrow-toprow)<=nvisrows);
    xx=pos_x+marginleft+lineWidth(cursorpos-cursorcol,cursorcol)-1;
    if(0<=xx+3 && xx-2<viewport_w){
      FXDCWindow dc(this);
      fh=font->getFontHeight();
      yt=pos_y+margintop+cursorrow*fh;
      yb=yt+fh-1;
      if(!(state&FLAG_CARET)){

        // Repaint cursor in background to erase it
        dc.setForeground(backColor);
        dc.drawLine(xx,yt,xx,yb);
        dc.drawLine(xx+1,yt,xx+1,yb);
        dc.drawLine(xx-2,yt,xx+3,yt);
        dc.drawLine(xx-2,yb,xx+3,yb);

        // Clip repaint of the text to the margins
        dc.setClipRectangle(marginleft,margintop,viewport_w-marginleft-marginright,viewport_h-margintop-marginbottom);

        // Restore text
        dc.setTextFont(font);
        drawTextRow(dc,cursorrow-toprow,xx-3,xx+3,0,INT_MAX);
        }
      else{
        dc.setForeground(cursorColor);
        dc.drawLine(xx,yt,xx,yb);
        dc.drawLine(xx+1,yt,xx+1,yb);
        dc.drawLine(xx-2,yt,xx+3,yt);
        dc.drawLine(xx-2,yb,xx+3,yb);
        }
      }
    }
  }


// Erase cursor overhang outside of margins
void FXText::eraseCursorOverhang(){
  register FXint xx,yt,yb,fh;
  if(visrows[0]<=cursorpos && cursorpos<=visrows[nvisrows]){
    FXASSERT(0<=cursorpos && cursorpos<=length);
    FXASSERT(0<=cursorrow && cursorrow<nrows);
    FXASSERT(0<=(cursorrow-toprow));
    FXASSERT((cursorrow-toprow)<=nvisrows);
    xx=pos_x+marginleft+lineWidth(cursorpos-cursorcol,cursorcol)-1;
    if(0<=xx+3 && xx-2<viewport_w){
      FXDCWindow dc(this);
      fh=font->getFontHeight();
      yt=pos_y+margintop+cursorrow*fh;
      yb=yt+fh-1;
      if(xx-2<=marginleft && 0<=xx+3){
        dc.setForeground(backColor);
        dc.fillRectangle(0,yt,marginleft,fh);
        }
      if(viewport_w-marginright<=xx+3 && xx-2<=viewport_w){
        dc.setForeground(backColor);
        dc.fillRectangle(viewport_w-marginright,yt,marginright,fh);
        }
      if(yt<=margintop && 0<=yb){
        dc.setForeground(backColor);
        dc.fillRectangle(xx-2,0,5,margintop);
        }
      if(viewport_h-marginbottom<=yb && yt<viewport_h){
        dc.setForeground(backColor);
        dc.fillRectangle(xx-2,viewport_h-marginbottom,5,marginbottom);
        }
      }
    }
  }


// Draw range of text
void FXText::drawTextRange(FXDCWindow& dc,FXint beg,FXint end){
  register FXint tl,bl,fc,lc,ln;
  if(beg<visrows[nvisrows] && visrows[0]<end){
    if(beg<visrows[0]) beg=visrows[0];
    if(end>visrows[nvisrows]) end=visrows[nvisrows];
    if(beg<end){
      tl=posToLine(beg,0);
      fc=beg-visrows[tl];
      bl=posToLine(end,tl);
      lc=end-visrows[bl];
      FXASSERT(0<=tl && tl<=bl && bl<=nvisrows);
      if(tl==bl){
        drawTextRow(dc,tl,0,width,fc,lc);
        }
      else{
        drawTextRow(dc,tl,0,width,fc,INT_MAX);
        for(ln=tl+1; ln<bl; ln++){
          drawTextRow(dc,ln,0,width,0,INT_MAX);
          }
        drawTextRow(dc,bl,0,width,0,lc);
        }
      }
    }
  }


// Repaint rectangle of text
void FXText::drawTextRectangle(FXDCWindow& dc,FXint x,FXint y,FXint w,FXint h){
  register FXint tl,bl,yy,hh,ln;
  FXASSERT(0<w && 0<h);
  hh=font->getFontHeight();
  yy=pos_y+margintop+toprow*hh;
  tl=(y-yy)/hh;
  bl=(y+h-yy)/hh;
  if(tl<0) tl=0;
  if(bl>=nvisrows) bl=nvisrows-1;
  if(tl>bl) return; 
  for(ln=tl; ln<=bl; ln++){
    drawTextRow(dc,ln,x,x+w,0,INT_MAX);
    }
  }


// Repaint text range
void FXText::updateRange(FXint beg,FXint end){
  register FXint tl,bl,fc,lc,ty,by,lx,rx,t;
  if(beg>end){t=beg;beg=end;end=t;}
  if(beg<visrows[nvisrows] && visrows[0]<end && beg<end){
    if(beg<visrows[0]) beg=visrows[0];
    if(end>visrows[nvisrows]) end=visrows[nvisrows];
    tl=posToLine(beg,0);
    fc=beg-visrows[tl];
    bl=posToLine(end,tl);
    lc=end-visrows[bl];
    if(tl==bl){
      ty=pos_y+margintop+(toprow+tl)*font->getFontHeight();
      by=ty+font->getFontHeight();
      lx=pos_x+marginleft+lineWidth(visrows[tl],fc);
      if(end<(visrows[tl+1]-1)) rx=pos_x+marginleft+lineWidth(visrows[tl],lc); else rx=width;
      }
    else{
      ty=pos_y+margintop+(toprow+tl)*font->getFontHeight();
      by=pos_y+margintop+(toprow+bl+1)*font->getFontHeight();
      lx=0;
      rx=width;
      }
    update(lx,ty,rx-lx,by-ty);
    }
  }


// Draw item list
long FXText::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  FXDCWindow dc(this,event);
  dc.setTextFont(font);
//dc.setForeground(FXRGB(255,0,0));
//dc.fillRectangle(event->rect.x,event->rect.y,event->rect.w,event->rect.h);
//dc.setForeground(backColor);
//dc.fillRectangle(event->rect.x,event->rect.y,event->rect.w,event->rect.h);
  dc.setClipRectangle(marginleft,margintop,viewport_w-marginright-marginleft,viewport_h-margintop-marginbottom);
  drawTextRectangle(dc,event->rect.x,event->rect.y,event->rect.w,event->rect.h);
  dc.setClipRectangle(0,0,width,height);
  dc.setForeground(backColor);
  dc.fillRectangle(0,0,viewport_w,margintop);
  dc.fillRectangle(0,margintop,marginleft,viewport_h-margintop-marginbottom);
  dc.fillRectangle(viewport_w-marginright,margintop,marginright,viewport_h-margintop-marginbottom);
  dc.fillRectangle(0,viewport_h-marginbottom,viewport_w,marginbottom);
  drawCursor(flags);
  return 1;
  }


/*******************************************************************************/


// Move the cursor
void FXText::setCursorPos(FXint pos){
  register FXint cursorstartnew,cursorstartold;
  if(pos>length) pos=length;
  if(pos<0) pos=0; 
  if(cursorpos!=pos){
    showCursor(0);
    cursorstartold=cursorpos-cursorcol;
    cursorstartnew=rowStart(pos);
    cursorcol=pos-cursorstartnew;
    cursorpos=pos;
    if(cursorstartnew<cursorstartold)
      cursorrow=cursorrow-countRows(cursorstartnew,cursorstartold);
    else
      cursorrow=cursorrow+countRows(cursorstartold,cursorstartnew);
    FXTRACE((150,"setCursorPos(%d): cursorpos=%d cursorrow=%d cursorcol=%d cursorstartold=%d cursorstartnew=%d\n",pos,cursorpos,cursorrow,cursorcol,cursorstartold,cursorstartnew));
    showCursor(FLAG_CARET);
    prefcol=-1;
    }
  }


// Set anchor position
void FXText::setAnchorPos(FXint pos){
  if(pos>length) pos=length;
  if(pos<0) pos=0; 
  anchorpos=pos;
  selstartpos=pos;
  selendpos=pos;
  }


// Kill the selection
FXbool FXText::killSelection(){
  if(hasSelection()){
    releaseSelection();
    updateRange(selstartpos,selendpos);
    selstartpos=anchorpos;
    selendpos=anchorpos;
    return TRUE;
    }
  return FALSE;
  }


// Set selection
FXbool FXText::setSelection(FXint pos,FXint len){
  setAnchorPos(pos);
  setCursorPos(pos+len);
  extendSelection(cursorpos,SELECT_CHARS);
  return TRUE;
  }


// Extend selection
FXbool FXText::extendSelection(FXint pos,FXTextSelectionMode mode){
  FXbool changes=FALSE;
  FXDragType types[1];
  FXint sp,ep;
  
  // Validate position
  if(pos<0) pos=0; 
  if(pos>length) pos=length;
  
  // Did position change?
  switch(mode){
 
    // Selecting words
    case SELECT_WORDS:
      if(pos<=anchorpos){
        sp=wordStart(pos);
        ep=wordEnd(anchorpos);
        }
      else{
        sp=wordStart(anchorpos);
        ep=wordEnd(pos);
        }
      break;
 
    // Selecting lines
    case SELECT_LINES:
      if(pos<=anchorpos){
        sp=rowStart(pos);
        ep=nextRow(anchorpos);
        }
      else{
        sp=rowStart(anchorpos);
        ep=nextRow(pos);
        }
      break;
 
    // Selecting characters
    default:
      if(pos<=anchorpos){
        sp=pos;
        ep=anchorpos;
        }
      else{
        sp=anchorpos;
        ep=pos;
        }
      break;
    }
  
  // Got a selection at all?
  if(sp!=ep){
    types[0]=stringType;
    acquireSelection(types,1);
    }
  else{
    releaseSelection();
    }
  
  // Start of selection changed
  if(sp!=selstartpos){
    updateRange(sp,selstartpos);
    selstartpos=sp;
    changes=TRUE;
    }
  
  // End of selection changed
  if(ep!=selendpos){
    updateRange(selendpos,ep);
    selendpos=ep;
    changes=TRUE;
    }
  
  // Must be true
  FXASSERT(0<=selstartpos);
  FXASSERT(selendpos<=length);
  FXASSERT(selstartpos<=selendpos);
  return changes;
  }

/*******************************************************************************/


// Change top margin
void FXText::setMarginTop(FXint mt){
  if(margintop!=mt){
    margintop=mt;
    recalc();
    update();
    }
  }

// Change bottom margin
void FXText::setMarginBottom(FXint mb){
  if(marginbottom!=mb){
    marginbottom=mb;
    recalc();
    update();
    }
  }


// Change left margin
void FXText::setMarginLeft(FXint ml){
  if(marginleft!=ml){
    marginleft=ml;
    recalc();
    update();
    }
  }


// Change right margin
void FXText::setMarginRight(FXint mr){
  if(marginright!=mr){
    marginright=mr;
    recalc();
    update();
    }
  }


// Set wrap columns
void FXText::setWrapColumns(FXint cols){
  if(cols!=wrapcolumns){
    wrapcolumns=cols;
    //if(id()&&(options&TEXT_FIXEDWRAP)){wrapwidth=wrapcolumns*font->getTextWidth(" ",1);}
    recalc();
    update();
    }
  }


// Set tab columns
void FXText::setTabColumns(FXint cols){
  if(cols!=tabcolumns){
    tabcolumns=cols;
    if(id()){tabwidth=tabcolumns*font->getTextWidth(" ",1);}
    recalc();
    update();
    }
  }


// Set text color
void FXText::setTextColor(FXColor clr){
  if(clr!=textColor){
    textColor=clr;
    update();
    }
  }


// Set select background color
void FXText::setSelBackColor(FXColor clr){
  if(clr!=selbackColor){
    selbackColor=clr;
    update();
    }
  }


// Set selected text color
void FXText::setSelTextColor(FXColor clr){
  if(clr!=seltextColor){
    seltextColor=clr;
    update();
    }
  }


// Set cursor color
void FXText::setCursorColor(FXColor clr){
  cursorColor=clr;
  }


// Change help text
void FXText::setHelpText(const FXString& text){
  help=text;
  }


// Change text style
void FXText::setTextStyle(FXuint style){
  FXuint opts=(options&~TEXT_MASK) | (style&TEXT_MASK);
  if(options!=opts){
    options=opts;
    recalc();
    update();
    }
  }
  
  
// Get text style
FXuint FXText::getTextStyle() const {
  return (options&TEXT_MASK); 
  }


// Return true if editable
FXbool FXText::isEditable() const { 
  return (options&TEXT_READONLY)==0; 
  }


// Set widget is editable or not
void FXText::setEditable(FXbool edit){
  if(edit) options&=~TEXT_READONLY; else options|=TEXT_READONLY;
  }


/*******************************************************************************/


// Set text color
long FXText::onCmdTextColor(FXObject*,FXSelector,void* ptr){
  setTextColor((FXColor)(long)ptr);
  return 1;
  }


// Update text color
long FXText::onUpdTextColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)textColor);
  return 1;
  }


// Set back color
long FXText::onCmdBackColor(FXObject*,FXSelector,void* ptr){
  setBackColor((FXColor)(long)ptr);
  return 1;
  }


// Set back color
long FXText::onUpdBackColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)backColor);
  return 1;
  }


// Set selection color
long FXText::onCmdSelTextColor(FXObject*,FXSelector,void* ptr){
  setSelTextColor((FXColor)(long)ptr);
  return 1;
  }


// Update selection color
long FXText::onUpdSelTextColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)seltextColor);
  return 1;
  }


// Set selection background color
long FXText::onCmdSelBackColor(FXObject*,FXSelector,void* ptr){
  setSelBackColor((FXColor)(long)ptr);
  return 1;
  }


// Update selection background color
long FXText::onUpdSelBackColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)selbackColor);
  return 1;
  }


// Set cursor color
long FXText::onCmdCursorColor(FXObject*,FXSelector,void* ptr){
  setCursorColor((FXColor)(long)ptr);
  return 1;
  }


// Update cursor color
long FXText::onUpdCursorColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)cursorColor);
  return 1;
  }


// Update value from a message
long FXText::onCmdSetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdSetStringValue: NULL pointer.\n",getClassName()); }
  setText(*((FXString*)ptr));
  return 1;
  }


// Obtain value from text 
long FXText::onCmdGetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetStringValue: NULL pointer.\n",getClassName()); }
  *((FXString*)ptr)=getText();
  return 1;
  }


/*******************************************************************************/


// Save object to stream
void FXText::save(FXStream& store) const {
  FXScrollArea::save(store);
  FXASSERT(0);//// FIXME
  }


// Load object from stream
void FXText::load(FXStream& store){
  FXScrollArea::load(store);
  FXASSERT(0);//// FIXME
  }


// Clean up
FXText::~FXText(){
  if(blinker) getApp()->removeTimeout(blinker);
  FXFREE(&buffer);
  FXFREE(&visrows);
  FXFREE(&clipbuffer);
  buffer=(FXchar*)-1;
  clipbuffer=(FXchar*)-1;
  visrows=(FXint*)-1;
  font=(FXFont*)-1;
  blinker=(FXTimer*)-1;
  }
