/********************************************************************************
*                                                                               *
*          B u f f e r - O v e r r u n   S a f e   F o r m a t t i n g          *
*                                                                               *
*********************************************************************************
* Copyright (C) 2000 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: fxsnprintf.cpp,v 1.7 2000/10/16 05:27:46 jeroen Exp $                    *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxpriv.h"

/*
  Notes:
  
  - These routines are based on the original routines in the Apache
    web server, which are themselves based on routines written by
    Panos Tsirigotis <panos@alumni.cs.colorado.edu> for xinetd.
    
  - The following notice must be included to with this code:

      The Apache Software License, Version 1.1

      Copyright (c) 2000 The Apache Software Foundation.  All rights
      reserved.

      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions
      are met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. The end-user documentation included with the redistribution,
         if any, must include the following acknowledgment:
            "This product includes software developed by the
             Apache Software Foundation (http://www.apache.org/)."
         Alternately, this acknowledgment may appear in the software itself,
         if and wherever such third-party acknowledgments normally appear.

      4. The names "Apache" and "Apache Software Foundation" must
         not be used to endorse or promote products derived from this
         software without prior written permission. For written
         permission, please contact apache@apache.org.

      5. Products derived from this software may not be called "Apache",
         nor may "Apache" appear in their name, without prior written
         permission of the Apache Software Foundation.

      THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
      WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
      OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
      DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
      ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
      SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
      LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
      USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
      OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
      OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
      SUCH DAMAGE.
      
  - Format strings recognized:
  
    % [flags] [width] [.precision] [size] format
    
      flags can be {-, +, #, ' ', 0}
      width is a sequence of digits, or *
      precision is a sequence of digits, or *
      size is one of {l, h, hh}
      format is one of {u, d, i, o, x, X, s, f, e, E, g, G, c, p}

  - Any bugs herein are probably are probably my fault.
  
  - Sources:

    cvt.c - IEEE floating point formatting routines for FreeBSD
    from GNU libc-4.6.27.  
    
  - No NaN's or Inf yet.
    The C99 standard specifies `[-]inf' or `[-]infinity' for infinity, and a 
    string starting with `nan' for NaN, in the case of f conversion, and `[-]INF' 
    or `[-]INFINITY' or `NAN*' in the case of F conversion. 
    
*/


#define FLOAT_DIGITS            6     // Number of float digits
#define EXPONENT_LENGTH        10     // Exponent length
#define LEFT                    0     // Left alignment
#define RIGHT                   1     // Right alignment
#define IS_LONG                 0     // Long size (l)
#define IS_SHORT                1     // Short size (h)
#define IS_INT                  2     // Default size
#define IS_CHAR                 3     // Byte size (hh)
#define NUM_BUF_SIZE          512     // Needed to convert numbers; don't change
#define NDIG                   80     // Maximum space used for numbers
#define NUM(c)          ((c)-'0')     // Number from ascii characters 0..9


/*
 * Structure used by the variable-formatter routines.
 */
struct apr_vformatter_buff_t {
  char *curpos;
  char *endpos;
  };


/*******************************************************************************/
  
static const char low_digits[]="0123456789abcdef";
static const char upper_digits[]="0123456789ABCDEF";

static const char S_NULL[]="(null)";
static const int  S_NULL_LEN=6;



// Converts to decimal
// the number of digits is specified by ndigit
// decpt is set to the position of the decimal point
// sign is set to 0 for positive, 1 for negative
static char *cvt(double arg,int ndigits,int *decpt,int *sign,int eflag,char *buf){
  register char *p,*p1;
  register int r2;
  double fi,fj;
  if(ndigits>=NDIG-1) ndigits=NDIG-2;           // Make sure stays inside buffer
  r2=0;
  *sign=0;
  p=&buf[0];
  if(arg<0){
    *sign=1;
    arg=-arg;
    }
  arg=modf(arg,&fi);
  p1=&buf[NDIG];
  if(fi!=0){                                    // Integer part
    p1=&buf[NDIG];
    while(fi!=0){
      fj=modf(fi/10,&fi);
      *--p1=(int)((fj+.03)*10)+'0';
      r2++;
      }
    while(p1<&buf[NDIG]) *p++=*p1++;
    }
  else if(arg>0){                               // Fractional part
    while((fj=arg*10)<1){
      arg=fj;
      r2--;
      }
    }
  p1=&buf[ndigits];
  if(eflag==0) p1+=r2;
  *decpt=r2;
  if(p1<&buf[0]){
    buf[0]='\0';
    return buf;
    }
  while(p<=p1 && p<&buf[NDIG]){
    arg*=10;
    arg=modf(arg,&fj);
    *p++=(int)fj+'0';
    }
  if(p1>=&buf[NDIG]){
    buf[NDIG-1]='\0';
    return buf;
    }
  p=p1;
  *p1+=5;
  while(*p1>'9'){
    *p1='0';
    if(p1>buf){
      ++ * --p1;
      }
    else{
      *p1='1';
      (*decpt)++;
      if(eflag==0){
        if(p>buf) *p='0';
        p++;
        }
      }
    }
  *p='\0';
  return buf;
  }


static char *ecvt(double arg,int ndigits,int *decpt,int *sign,char *buf){
  return cvt(arg,ndigits,decpt,sign,1,buf);
  }


static char *fcvt(double arg,int ndigits,int *decpt,int *sign,char *buf){
  return cvt(arg,ndigits,decpt,sign,0,buf);
  }


// Floating output conversion to minimal length string
static char *gcvt(double number,int ndigit,char *buf,int altform){
  char buf1[NDIG],*p1,*p2;
  int  sign,decpt,i;
  p1=ecvt(number,ndigit,&decpt,&sign,buf1);
  p2=buf;
  if(sign) *p2++='-';
  for(i=ndigit-1; i>0 && p1[i]=='0'; i--) ndigit--;
  if((decpt>=0 && decpt-ndigit>4) || (decpt<0 && decpt<-3)){      // use E-style 
    decpt--;
    *p2++=*p1++;
    *p2++='.';
    for(i=1; i<ndigit; i++) *p2++=*p1++;
    *p2++='e';
    if(decpt<0){
      decpt=-decpt;
      *p2++='-';
      }
    else{
      *p2++='+';
      }
    if(decpt/100>0) *p2++=decpt/100+'0';
    if(decpt/10>0) *p2++=(decpt%100)/10+'0';
    *p2++=decpt%10+'0';
    }
  else{
    if(decpt<=0){
      if(*p1!='0') *p2++='.';
      while(decpt<0){
        decpt++;
        *p2++='0';
        }
      }
    for(i=1; i<=ndigit; i++){
      *p2++=*p1++;
      if(i==decpt) *p2++='.';
      }
    if(ndigit<decpt){
      while(ndigit++<decpt) *p2++='0';
      *p2++='.';
      }
    }
  if(p2[-1]=='.' && !altform) p2--;
  *p2='\0';
  return buf;
  }


// Convert num to a base X number where X is a power of 2. nbits determines X.
// For example, if nbits is 3, we do base 8 conversion
// Return value:
//      a pointer to a string containing the number
//
// The caller provides a buffer for the string: that is the buf_end argument
// which is a pointer to the END of the buffer + 1 (i.e. if the buffer
// is declared as buf[ 100 ], buf_end should be &buf[ 100 ])
static char *conv_p2(unsigned long num,int nbits,char format,char *buf_end,int *len){
  register const char *digits=(format=='X')?upper_digits:low_digits;
  register unsigned long mask=(1<<nbits)-1;
  register char *p=buf_end;
  do{
    *--p=digits[num&mask];
    num>>=nbits;
    }
  while(num);
  *len=buf_end-p;
  return p;
  }


// Convert num to its decimal format.
// Return value:
//   - a pointer to a string containing the number (no sign)
//   - len contains the length of the string
//   - is_negative is set to TRUE or FALSE depending on the sign
//     of the number (always set to FALSE if is_unsigned is TRUE)
//
// The caller provides a buffer for the string: that is the buf_end argument
// which is a pointer to the END of the buffer + 1 (i.e. if the buffer
// is declared as buf[ 100 ], buf_end should be &buf[ 100 ])
static char *conv_10(long num,int is_unsigned,int *is_negative,char *buf_end,int *len){
  register unsigned long magnitude,new_magnitude;
  register char *p=buf_end;
  if(is_unsigned){
    magnitude=(unsigned long)num;
    *is_negative=FALSE;
    }
  else{
    // On a 2's complement machine, negating the most negative integer
    // results in a number that cannot be represented as a signed integer.
    // Here is what we do to obtain the number's magnitude:
    //      a. add 1 to the number
    //      b. negate it (becomes positive)
    //      c. convert it to unsigned
    //      d. add 1
    if(num<0){
      long t=num+1;
      magnitude=((unsigned long)-t)+1;
      *is_negative=1;
      }
    else{
      magnitude=(unsigned long)num;
      *is_negative=0;
      }
    }

  // We use a do-while loop so that we write at least 1 digit 
  do{
    new_magnitude=magnitude/10;
    *--p=(char)(magnitude-new_magnitude*10+'0');
    magnitude=new_magnitude;
    }
  while(magnitude);
  *len=buf_end-p;
  return p;
  }


// Convert a floating point number to a string formats 'f', 'e' or 'E'.
// The result is placed in buf, and len denotes the length of the string
// The sign is returned in the is_negative argument (and is not placed in buf).
static char *conv_fp(char format,double num,int add_dp,int precision,int *is_negative,char *buf,int *len){
  int decimal_point,t_len,exponent_is_negative;
  char buf1[NDIG],temp[EXPONENT_LENGTH];
  register char *s=buf;
  register char *p;

  // F format
  if(format=='f'){
    p=fcvt(num,precision,&decimal_point,is_negative,buf1);
    }
  
  // either e or E format
  else{
    p=ecvt(num,precision+1,&decimal_point,is_negative,buf1);
    }

  // Check for Infinity and NaN
  if(isalpha(*p)){
    *len=strlen(strcpy(buf,p));
    *is_negative=FALSE;
    return buf;
    }
  if(format=='f'){
    if(decimal_point<=0){
      *s++='0';
      if(precision>0){
        *s++='.';
        while(decimal_point++<0) *s++='0';
        }
      else{
        if(add_dp) *s++='.';
        }
      }
    else{
      while(decimal_point-->0) *s++=*p++;
      if(precision>0 || add_dp) *s++='.';
      }
    }
  else{
    *s++=*p++;
    if(precision>0 || add_dp) *s++='.';
    }

  // Copy the rest of p, the '\0' is NOT copied
  while(*p) *s++=*p++;

  if(format!='f'){
    *s++=format;                // either e or E
    decimal_point--;
    if(decimal_point!=0){
      p=conv_10(decimal_point,FALSE,&exponent_is_negative,&temp[EXPONENT_LENGTH],&t_len);
      *s++=exponent_is_negative?'-':'+';

      // Make sure the exponent has at least 2 digits
      if(t_len==1) *s++='0';
      while(t_len--) *s++=*p++;
      }
    else{
      *s++='+';
      *s++='0';
      *s++='0';
      }
    }

  *len=s-buf;
  return buf;
  }


// Do format conversion placing the output in buffer
int apr_vformatter(apr_vformatter_buff_t *vbuff,const char *fmt,va_list ap){
  int alternate_form,print_sign,print_blank,adjust_precision,adjust_width;
  int precision,adjust,min_width,s_len,var_type,is_negative,i,cc;
  char num_buf[NUM_BUF_SIZE],char_buf[2],pad_char,prefix_char;
  char *sp,*bep,*s,*q;
  unsigned long ui_num;
  double fp_num;
  long i_num;
  
  cc=0;
  sp=vbuff->curpos;
  bep=vbuff->endpos;

  while(*fmt){
    
    // Normal character
    if(*fmt!='%'){
      if(sp<bep) *sp++=*fmt;
      cc++;
      }

    // Format character
    else{

      // Default variable settings
      adjust=RIGHT;
      alternate_form=FALSE;
      print_sign=FALSE;
      print_blank=FALSE;
      pad_char=' ';
      prefix_char='\0';
      adjust_width=FALSE;
      adjust_precision=FALSE;

      fmt++;

      // See if flags, width, or precision have been specified
      if(!islower(*fmt)){
        
        // Parse flag characters #, 0, -, ' ', +
        for(;; fmt++){
          if(*fmt=='-') adjust=LEFT;              // Align on left
          else if(*fmt=='+') print_sign=TRUE;     // Always print sign
          else if(*fmt=='#') alternate_form=TRUE; // Alternate form
          else if(*fmt==' ') print_blank=TRUE;    // Print blank if no sign
          else if(*fmt=='0') pad_char='0';        // Pad with zeroes
          else break;
          }

        // Field width
        if(isdigit(*fmt)){
          min_width=NUM(*fmt++);
          while(isdigit(*fmt)){
            min_width*=10;
            min_width+=NUM(*fmt++);
            }
          adjust_width=TRUE;
          }
        else if(*fmt=='*'){
          min_width=va_arg(ap,int);
          fmt++;
          if(min_width<0){
            adjust=LEFT;
            min_width=-min_width;
            }
          adjust_width=TRUE;
          }

        // Precision
        if(*fmt=='.'){
          fmt++;
          if(isdigit(*fmt)){
            precision=NUM(*fmt++);
            while(isdigit(*fmt)){
              precision*=10;
              precision+=NUM(*fmt++);
              }
            if(precision>100) precision=100;      // Keep in check
            }
          else if(*fmt=='*'){
            precision=va_arg(ap,int);
            fmt++;
            if(precision<0) precision=0;
            if(precision>100) precision=100;      // Keep in check
            }
          else{
            precision=0;
            }
          adjust_precision=TRUE;
          }
        }
      
      // Check modifiers
      if(*fmt=='l'){
        var_type=IS_LONG;
        fmt++;
        }
      else if(*fmt=='h'){
        var_type=IS_SHORT;
        fmt++;
        if(*fmt=='h'){
          var_type=IS_CHAR;
          fmt++;
          }
        }
      else{
        var_type=IS_INT;
        }

      // Argument extraction and printing.
      // First we determine the argument type.
      // Then, we convert the argument to a string.
      // On exit from the switch, s points to the string that
      // must be printed, s_len has the length of the string
      // The precision requirements, if any, are reflected in s_len.
      // The variable pad_char is reset to ' ' by non-numeric formats.
      switch(*fmt){
        case 'u':
          if(var_type==IS_LONG) 
            i_num=(long)va_arg(ap,unsigned long);
          else if(var_type==IS_SHORT) 
            i_num=(long)(unsigned short)va_arg(ap,unsigned int);
          else if(var_type==IS_CHAR)
            i_num=(long)(unsigned char)va_arg(ap,unsigned int);
          else 
            i_num=(long)va_arg(ap,unsigned int);
          s=conv_10(i_num,1,&is_negative,&num_buf[NUM_BUF_SIZE],&s_len);
          if(adjust_precision){
            while(s_len<precision){ *--s='0'; s_len++; }
            }
          break;

        case 'd':
        case 'i':
          if(var_type==IS_LONG) 
            i_num=(long)va_arg(ap,long);
          else if(var_type==IS_SHORT) 
            i_num=(long)(short)va_arg(ap,int);
          else if(var_type==IS_CHAR)
            i_num=(long)(signed char)va_arg(ap,int);
          else 
            i_num=(long)va_arg(ap,int);
          s=conv_10(i_num,0,&is_negative,&num_buf[NUM_BUF_SIZE],&s_len);
          if(adjust_precision){
            while(s_len<precision){ *--s='0'; s_len++; }
            }
          if(is_negative) prefix_char='-';
          else if(print_sign) prefix_char='+';
          else if(print_blank) prefix_char=' ';
          break;

        case 'o':
          if(var_type==IS_LONG) 
            ui_num=(unsigned long)va_arg(ap,unsigned long);
          else if(var_type==IS_SHORT) 
            ui_num=(unsigned long)(unsigned short)va_arg(ap,unsigned int);
          else if(var_type==IS_CHAR)
            i_num=(unsigned long)(unsigned char)va_arg(ap,unsigned int);
          else 
            ui_num=(unsigned long)va_arg(ap,unsigned int);
          s=conv_p2(ui_num,3,*fmt,&num_buf[NUM_BUF_SIZE],&s_len);
          if(adjust_precision){
            while(s_len<precision){ *--s='0'; s_len++; }
            }
          if(alternate_form && *s!='0'){
            *--s='0';
            s_len++;
            }
          break;

        case 'x':
        case 'X':
          if(var_type==IS_LONG) 
            ui_num=(unsigned long)va_arg(ap,unsigned long);
          else if(var_type==IS_SHORT) 
            ui_num=(unsigned long)(unsigned short)va_arg(ap,unsigned int);
          else if(var_type==IS_CHAR)
            i_num=(unsigned long)(unsigned char)va_arg(ap,unsigned int);
          else 
            ui_num=(unsigned long)va_arg(ap,unsigned int);
          s=conv_p2(ui_num,4,*fmt,&num_buf[NUM_BUF_SIZE],&s_len);
          if(adjust_precision){
            while(s_len<precision){ *--s='0'; s_len++; }
            }
          if(alternate_form && ui_num!=0){
            *--s=*fmt;        // 'x' or 'X'
            *--s='0';
            s_len+=2;
            }
          break;

        case 's':
          s=va_arg(ap,char*);
          if(s!=NULL){
            s_len=strlen(s);
            if(adjust_precision && precision<s_len) s_len=precision;
            }
          else{
            s=(char*)S_NULL;
            s_len=S_NULL_LEN;
            }
          pad_char=' ';       // Non-numeric stuff padded with spaces
          break;

        case 'f':
        case 'e':
        case 'E':
          fp_num=va_arg(ap,double);
          // We use &num_buf[ 1 ], so that we have room for the sign
          s=conv_fp(*fmt,fp_num,alternate_form,(adjust_precision==FALSE) ? FLOAT_DIGITS : precision,&is_negative,&num_buf[1],&s_len);
          if(is_negative) prefix_char='-';
          else if(print_sign) prefix_char='+';
          else if(print_blank) prefix_char=' ';
          break;

        case 'g':
        case 'G':
          if(adjust_precision==FALSE) precision=FLOAT_DIGITS;
          else if(precision==0) precision=1;
          // We use &num_buf[ 1 ], so that we have room for the sign
          s=gcvt(va_arg(ap,double),precision,&num_buf[1],alternate_form);
          if(*s=='-') prefix_char=*s++;
          else if(print_sign) prefix_char='+';
          else if(print_blank) prefix_char=' ';
          s_len=strlen(s);
          if(alternate_form && (q=strchr(s,'.'))==NULL){
            s[s_len++]='.';
            s[s_len]='\0';    // delimit for following strchr()
            }
          if(*fmt=='G' && (q=strchr(s,'e'))!=NULL) *q='E';
          break;

        case 'c':
          char_buf[0]=(char)(va_arg(ap,int));
          s=&char_buf[0];
          s_len=1;
          pad_char=' ';       // Non-numeric stuff padded with spaces
          break;

        case '%':
          char_buf[0]='%';
          s=&char_buf[0];
          s_len=1;
          pad_char=' ';       // Non-numeric stuff padded with spaces
          break;

        case 'n':
          if(var_type==IS_LONG) 
            *(va_arg(ap,long*))=cc;
          else if(var_type==IS_SHORT) 
            *(va_arg(ap,short*))=cc;
          else if(var_type==IS_CHAR) 
            *(va_arg(ap,char*))=cc;
          else 
            *(va_arg(ap,int*))=cc;
          break;

        case 'p':
          ui_num=(unsigned long)va_arg(ap,void*);
          s=conv_p2(ui_num,4,'x',&num_buf[NUM_BUF_SIZE],&s_len);
          pad_char=' ';       // Non-numeric stuff padded with spaces
          break;

        case '\0':            // The last character of the format string was %
          continue;

        default:              // Unrecognized option
          char_buf[0]='%';
          char_buf[1]=*fmt;
          s=char_buf;
          s_len=2;
          pad_char=' ';       // Non-numeric stuff padded with spaces
          break;
        }

      if(prefix_char!='\0' && s!=S_NULL && s!=char_buf){
        *--s=prefix_char;
        s_len++;
        }

      // Pad to the right
      if(adjust_width && adjust==RIGHT && min_width>s_len){
        if(pad_char=='0' && prefix_char!='\0'){
          if(sp<bep) *sp++=*s;
          cc++;
          s++;
          s_len--;
          min_width--;
          }
        do{
          if(sp<bep) *sp++=pad_char;
          cc++;
          min_width--;
          }
         while(min_width>s_len);
        }
      
      // Output the string s
      for(i=s_len; i!=0; i--){
        if(sp<bep) *sp++=*s;
        cc++;
        s++;
        }

      // Pad to the left
      if(adjust_width && adjust==LEFT && min_width>s_len){
        do{
          if(sp<bep) *sp++=pad_char;
          cc++;
          min_width--;
          }
         while(min_width>s_len);
        }
      }
    fmt++;
    } 
  vbuff->curpos=sp;
  return cc;
  }


int fxsnprintf(char *buf,int len,const char *format,...){
  int cc;
  va_list ap;
  apr_vformatter_buff_t vbuff;
  vbuff.curpos=buf;
  vbuff.endpos=buf+len-1;                 // save one byte for nul terminator
  va_start(ap,format);
  cc=apr_vformatter(&vbuff,format,ap);
  va_end(ap);
  if(0<len) *vbuff.curpos='\0';
  return cc;
  }


int fxvsnprintf(char *buf,int len,const char *format,va_list ap){
  int cc;
  apr_vformatter_buff_t vbuff;
  vbuff.curpos=buf;
  vbuff.endpos=buf+len-1;                 // save one byte for nul terminator
  cc=apr_vformatter(&vbuff,format,ap);
  if(0<len) *vbuff.curpos='\0';
  return cc;
  }
